"""
Manage grains on the minion
===========================

This state allows for grains to be set.

Grains set or altered with this module are stored in the 'grains'
file on the minions, By default, this file is located at: ``/etc/salt/grains``

.. note::
   This does **NOT** override any grains set in the minion config file.
"""

import re

from salt.defaults import DEFAULT_TARGET_DELIM


def exists(name, delimiter=DEFAULT_TARGET_DELIM):
    """
    Ensure that a grain is set

    name
        The grain name

    delimiter
        A delimiter different from the default can be provided.

    Check whether a grain exists. Does not attempt to check or set the value.
    """
    name = re.sub(delimiter, DEFAULT_TARGET_DELIM, name)
    ret = {"name": name, "changes": {}, "result": True, "comment": "Grain exists"}
    _non_existent = object()
    existing = __salt__["grains.get"](name, _non_existent)
    if existing is _non_existent:
        ret["result"] = False
        ret["comment"] = "Grain does not exist"
    return ret


def make_hashable(list_grain, result=None):
    """
    Ensure that a list grain is hashable.

    list_grain
        The list grain that should be hashable

    result
        This function is recursive, so it must be possible to use a
        sublist as parameter to the function. Should not be used by a caller
        outside of the function.

    Make it possible to compare two list grains to each other if the list
    contains complex objects.
    """
    result = result or set()
    for sublist in list_grain:
        if type(sublist) == list:
            make_hashable(sublist, result)
        else:
            result.add(frozenset(sublist))
    return result


def present(name, value, delimiter=DEFAULT_TARGET_DELIM, force=False):
    """
    Ensure that a grain is set

    .. versionchanged:: 2015.8.2

    name
        The grain name

    value
        The value to set on the grain

    force
        If force is True, the existing grain will be overwritten
        regardless of its existing or provided value type. Defaults to False

        .. versionadded:: 2015.8.2

    delimiter
        A delimiter different from the default can be provided.

        .. versionadded:: 2015.8.2

    It is now capable to set a grain to a complex value (ie. lists and dicts)
    and supports nested grains as well.

    If the grain does not yet exist, a new grain is set to the given value. For
    a nested grain, the necessary keys are created if they don't exist. If
    a given key is an existing value, it will be converted, but an existing value
    different from the given key will fail the state.

    If the grain with the given name exists, its value is updated to the new
    value unless its existing or provided value is complex (list or dict). Use
    `force: True` to overwrite.

    .. code-block:: yaml

      cheese:
        grains.present:
          - value: edam

      nested_grain_with_complex_value:
        grains.present:
          - name: icinga:Apache SSL
          - value:
            - command: check_https
            - params: -H localhost -p 443 -S

      with,a,custom,delimiter:
        grains.present:
          - value: yay
          - delimiter: ','
    """
    name = re.sub(delimiter, DEFAULT_TARGET_DELIM, name)
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    _non_existent = object()
    existing = __salt__["grains.get"](name, _non_existent)
    if existing == value:
        ret["comment"] = "Grain is already set"
        return ret
    if __opts__["test"]:
        ret["result"] = None
        if existing is _non_existent:
            ret["comment"] = f"Grain {name} is set to be added"
            ret["changes"] = {"new": name}
        else:
            ret["comment"] = f"Grain {name} is set to be changed"
            ret["changes"] = {"changed": {name: value}}
        return ret
    ret = __salt__["grains.set"](name, value, force=force)
    if ret["result"] is True and ret["changes"] != {}:
        ret["comment"] = f"Set grain {name} to {value}"
    ret["name"] = name
    return ret


def list_present(name, value, delimiter=DEFAULT_TARGET_DELIM):
    """
    .. versionadded:: 2014.1.0

    Ensure the value is present in the list-type grain. Note: If the grain that is
    provided in ``name`` is not present on the system, this new grain will be created
    with the corresponding provided value.

    name
        The grain name.

    value
        The value is present in the list type grain.

    delimiter
        A delimiter different from the default ``:`` can be provided.

        .. versionadded:: 2015.8.2

    The grain should be `list type <http://docs.python.org/2/tutorial/datastructures.html#data-structures>`_

    .. code-block:: yaml

        roles:
          grains.list_present:
            - value: web

    For multiple grains, the syntax looks like:

    .. code-block:: yaml

        roles:
          grains.list_present:
            - value:
              - web
              - dev
    """
    name = re.sub(delimiter, DEFAULT_TARGET_DELIM, name)
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    grain = __salt__["grains.get"](name)
    if grain:
        # check whether grain is a list
        if not isinstance(grain, list):
            ret["result"] = False
            ret["comment"] = f"Grain {name} is not a valid list"
            return ret
        if isinstance(value, list):
            if make_hashable(value).issubset(
                make_hashable(__salt__["grains.get"](name))
            ):
                ret["comment"] = f"Value {value} is already in grain {name}"
                return ret
            elif name in __context__.get("pending_grains", {}):
                # elements common to both
                intersection = set(value).intersection(
                    __context__.get("pending_grains", {})[name]
                )
                if intersection:
                    value = list(
                        set(value).difference(__context__["pending_grains"][name])
                    )
                    ret["comment"] = (
                        'Removed value {} from update due to context found in "{}".\n'.format(
                            value, name
                        )
                    )
            if "pending_grains" not in __context__:
                __context__["pending_grains"] = {}
            if name not in __context__["pending_grains"]:
                __context__["pending_grains"][name] = set()
            __context__["pending_grains"][name].update(value)
        else:
            if value in grain:
                ret["comment"] = f"Value {value} is already in grain {name}"
                return ret
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = "Value {1} is set to be appended to grain {0}".format(
                name, value
            )
            ret["changes"] = {"new": grain}
            return ret

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f"Grain {name} is set to be added"
        ret["changes"] = {"new": grain}
        return ret
    new_grains = __salt__["grains.append"](name, value)
    if isinstance(value, list):
        if not set(value).issubset(set(__salt__["grains.get"](name))):
            ret["result"] = False
            ret["comment"] = f"Failed append value {value} to grain {name}"
            return ret
    else:
        if value not in __salt__["grains.get"](name, delimiter=DEFAULT_TARGET_DELIM):
            ret["result"] = False
            ret["comment"] = f"Failed append value {value} to grain {name}"
            return ret
    ret["comment"] = f"Append value {value} to grain {name}"
    ret["changes"] = {"new": new_grains}
    return ret


def list_absent(name, value, delimiter=DEFAULT_TARGET_DELIM):
    """
    Delete a value from a grain formed as a list.

    .. versionadded:: 2014.1.0

    name
        The grain name.

    value
       The value to delete from the grain list.

    delimiter
        A delimiter different from the default ``:`` can be provided.

        .. versionadded:: 2015.8.2

    The grain should be `list type <http://docs.python.org/2/tutorial/datastructures.html#data-structures>`_

    .. code-block:: yaml

        roles:
          grains.list_absent:
            - value: db

    For multiple grains, the syntax looks like:

    .. code-block:: yaml

        roles:
          grains.list_absent:
            - value:
              - web
              - dev
    """

    name = re.sub(delimiter, DEFAULT_TARGET_DELIM, name)
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    comments = []
    grain = __salt__["grains.get"](name, None)
    if grain:
        if isinstance(grain, list):
            if not isinstance(value, list):
                value = [value]
            for val in value:
                if val not in grain:
                    comments.append(f"Value {val} is absent from grain {name}")
                elif __opts__["test"]:
                    ret["result"] = None
                    comments.append(f"Value {val} in grain {name} is set to be deleted")
                    if "deleted" not in ret["changes"]:
                        ret["changes"] = {"deleted": []}
                    ret["changes"]["deleted"].append(val)
                elif val in grain:
                    __salt__["grains.remove"](name, val)
                    comments.append(f"Value {val} was deleted from grain {name}")
                    if "deleted" not in ret["changes"]:
                        ret["changes"] = {"deleted": []}
                    ret["changes"]["deleted"].append(val)
            ret["comment"] = "\n".join(comments)
            return ret
        else:
            ret["result"] = False
            ret["comment"] = f"Grain {name} is not a valid list"
    else:
        ret["comment"] = f"Grain {name} does not exist"
    return ret


def absent(name, destructive=False, delimiter=DEFAULT_TARGET_DELIM, force=False):
    """
    .. versionadded:: 2014.7.0

    Delete a grain from the grains config file

    name
        The grain name

    destructive
        If destructive is True, delete the entire grain. If
        destructive is False, set the grain's value to None. Defaults to False.

    force
        If force is True, the existing grain will be overwritten
        regardless of its existing or provided value type. Defaults to False

        .. versionadded:: 2015.8.2

    delimiter
        A delimiter different from the default can be provided.

        .. versionadded:: 2015.8.2

    .. versionchanged:: 2015.8.2

    This state now support nested grains and complex values. It is also more
    conservative: if a grain has a value that is a list or a dict, it will
    not be removed unless the `force` parameter is True.

    .. code-block:: yaml

      grain_name:
        grains.absent
    """

    _non_existent = object()

    name = re.sub(delimiter, DEFAULT_TARGET_DELIM, name)
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    grain = __salt__["grains.get"](name, _non_existent)
    if grain is None:
        if __opts__["test"]:
            ret["result"] = None
            if destructive is True:
                ret["comment"] = f"Grain {name} is set to be deleted"
                ret["changes"] = {"deleted": name}
            return ret
        ret = __salt__["grains.set"](name, None, destructive=destructive, force=force)
        if ret["result"]:
            if destructive is True:
                ret["comment"] = f"Grain {name} was deleted"
                ret["changes"] = {"deleted": name}
        ret["name"] = name
    elif grain is not _non_existent:
        if __opts__["test"]:
            ret["result"] = None
            if destructive is True:
                ret["comment"] = f"Grain {name} is set to be deleted"
                ret["changes"] = {"deleted": name}
            else:
                ret["comment"] = f"Value for grain {name} is set to be deleted (None)"
                ret["changes"] = {"grain": name, "value": None}
            return ret
        ret = __salt__["grains.set"](name, None, destructive=destructive, force=force)
        if ret["result"]:
            if destructive is True:
                ret["comment"] = f"Grain {name} was deleted"
                ret["changes"] = {"deleted": name}
            else:
                ret["comment"] = f"Value for grain {name} was set to None"
                ret["changes"] = {"grain": name, "value": None}
        ret["name"] = name
    else:
        ret["comment"] = f"Grain {name} does not exist"
    return ret


def append(name, value, convert=False, delimiter=DEFAULT_TARGET_DELIM):
    """
    .. versionadded:: 2014.7.0

    Append a value to a list in the grains config file. The grain that is being
    appended to (name) must exist before the new value can be added.

    name
        The grain name

    value
        The value to append

    convert
        If convert is True, convert non-list contents into a list.
        If convert is False and the grain contains non-list contents, an error
        is given. Defaults to False.

    delimiter
        A delimiter different from the default can be provided.

        .. versionadded:: 2015.8.2

    .. code-block:: yaml

      grain_name:
        grains.append:
          - value: to_be_appended
    """
    name = re.sub(delimiter, DEFAULT_TARGET_DELIM, name)
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    grain = __salt__["grains.get"](name, None)

    # Check if bool(grain) is False or if the grain is specified in the minions
    # grains. Grains can be set to a None value by omitting a value in the
    # definition.
    if grain or name in __grains__:
        if isinstance(grain, list):
            if value in grain:
                ret["comment"] = (
                    f"Value {value} is already in the list for grain {name}"
                )
                return ret
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = "Value {1} in grain {0} is set to be added".format(
                    name, value
                )
                ret["changes"] = {"added": value}
                return ret
            __salt__["grains.append"](name, value)
            ret["comment"] = f"Value {value} was added to grain {name}"
            ret["changes"] = {"added": value}
        else:
            if convert is True:
                if __opts__["test"]:
                    ret["result"] = None
                    ret["comment"] = (
                        "Grain {} is set to be converted "
                        "to list and value {} will be "
                        "added".format(name, value)
                    )
                    ret["changes"] = {"added": value}
                    return ret
                grain = [] if grain is None else [grain]
                grain.append(value)
                __salt__["grains.setval"](name, grain)
                ret["comment"] = f"Value {value} was added to grain {name}"
                ret["changes"] = {"added": value}
            else:
                ret["result"] = False
                ret["comment"] = f"Grain {name} is not a valid list"
    else:
        ret["result"] = False
        ret["comment"] = f"Grain {name} does not exist"
    return ret



"""
Management of PostgreSQL extensions
===================================

A module used to install and manage PostgreSQL extensions.

.. code-block:: yaml

    adminpack:
      postgres_extension.present


.. versionadded:: 2014.7.0
"""

import logging

from salt.modules import postgres

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if the postgres module is present
    """
    if "postgres.create_extension" not in __salt__:
        return (
            False,
            "Unable to load postgres module.  Make sure `postgres.bins_dir` is set.",
        )
    return True


def present(
    name,
    if_not_exists=None,
    schema=None,
    ext_version=None,
    from_version=None,
    user=None,
    maintenance_db=None,
    db_user=None,
    db_password=None,
    db_host=None,
    db_port=None,
):
    """
    Ensure that the named extension is present.

    .. note::

        Before you can use the state to load an extension into a database, the
        extension's supporting files must be already installed.

    For more information about all of these options see ``CREATE EXTENSION`` SQL
    command reference in the PostgreSQL documentation.

    name
        The name of the extension to be installed

    if_not_exists
        Add an ``IF NOT EXISTS`` parameter to the DDL statement

    schema
        Schema to install the extension into

    ext_version
        Version to install

    from_version
        Old extension version if already installed

    user
        System user all operations should be performed on behalf of

    maintenance_db
        Database to act on

    db_user
        Database username if different from config or default

    db_password
        User password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {
        "name": name,
        "changes": {},
        "result": True,
        "comment": f"Extension {name} is already present",
    }
    db_args = {
        "maintenance_db": maintenance_db,
        "runas": user,
        "user": db_user,
        "password": db_password,
        "host": db_host,
        "port": db_port,
    }
    # check if extension exists
    mode = "create"
    mtdata = __salt__["postgres.create_metadata"](
        name, schema=schema, ext_version=ext_version, **db_args
    )

    # The extension is not present, install it!
    toinstall = postgres._EXTENSION_NOT_INSTALLED in mtdata
    if toinstall:
        mode = "install"
    toupgrade = False
    if postgres._EXTENSION_INSTALLED in mtdata:
        for flag in [postgres._EXTENSION_TO_MOVE, postgres._EXTENSION_TO_UPGRADE]:
            if flag in mtdata:
                toupgrade = True
                mode = "upgrade"
    cret = None
    if toinstall or toupgrade:
        if __opts__["test"]:
            ret["result"] = None
            if mode:
                ret["comment"] = "Extension {} is set to be {}ed".format(
                    name, mode
                ).replace("eed", "ed")
            return ret

        cret = __salt__["postgres.create_extension"](
            name=name,
            if_not_exists=if_not_exists,
            schema=schema,
            ext_version=ext_version,
            from_version=from_version,
            **db_args,
        )
    if cret:
        if mode.endswith("e"):
            suffix = "d"
        else:
            suffix = "ed"
        ret["comment"] = f"The extension {name} has been {mode}{suffix}"
        ret["changes"][name] = f"{mode.capitalize()}{suffix}"
    elif cret is not None:
        ret["comment"] = f"Failed to {mode} extension {name}"
        ret["result"] = False

    return ret


def absent(
    name,
    if_exists=None,
    restrict=None,
    cascade=None,
    user=None,
    maintenance_db=None,
    db_user=None,
    db_password=None,
    db_host=None,
    db_port=None,
):
    """
    Ensure that the named extension is absent.

    name
        Extension name of the extension to remove

    if_exists
        Add if exist slug

    restrict
        Add restrict slug

    cascade
        Drop on cascade

    user
        System user all operations should be performed on behalf of

    maintenance_db
        Database to act on

    db_user
        Database username if different from config or default

    db_password
        User password if any password for a specified user

    db_host
        Database host if different from config or default

    db_port
        Database port if different from config or default
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    db_args = {
        "maintenance_db": maintenance_db,
        "runas": user,
        "host": db_host,
        "user": db_user,
        "port": db_port,
        "password": db_password,
    }
    # check if extension exists and remove it
    exists = __salt__["postgres.is_installed_extension"](name, **db_args)
    if exists:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Extension {name} is set to be removed"
            return ret
        if __salt__["postgres.drop_extension"](
            name, if_exists=if_exists, restrict=restrict, cascade=cascade, **db_args
        ):
            ret["comment"] = f"Extension {name} has been removed"
            ret["changes"][name] = "Absent"
            return ret
        else:
            ret["result"] = False
            ret["comment"] = f"Extension {name} failed to be removed"
            return ret
    else:
        ret["comment"] = "Extension {} is not present, so it cannot be removed".format(
            name
        )

    return ret



"""
Test case for the vault SDB module
"""

import pytest

import salt.exceptions
import salt.sdb.vault as vault
import salt.utils.vault as vaultutil
from tests.support.mock import ANY, patch


@pytest.fixture
def configure_loader_modules():
    return {vault: {}}


@pytest.fixture
def data():
    return {"bar": "super awesome"}


@pytest.fixture
def read_kv(data):
    with patch("salt.utils.vault.read_kv", autospec=True) as read:
        read.return_value = data
        yield read


@pytest.fixture
def read_kv_not_found(read_kv):
    read_kv.side_effect = vaultutil.VaultNotFoundError


@pytest.fixture
def read_kv_not_found_once(read_kv, data):
    read_kv.side_effect = (vaultutil.VaultNotFoundError, data)
    yield read_kv


@pytest.fixture
def read_kv_err(read_kv):
    read_kv.side_effect = vaultutil.VaultPermissionDeniedError("damn")
    yield read_kv


@pytest.fixture
def write_kv():
    with patch("salt.utils.vault.write_kv", autospec=True) as write:
        yield write


@pytest.fixture
def write_kv_err(write_kv):
    write_kv.side_effect = vaultutil.VaultPermissionDeniedError("damn")
    yield write_kv


@pytest.mark.parametrize(
    "key,exp_path",
    [
        ("sdb://myvault/path/to/foo/bar", "path/to/foo"),
        ("sdb://myvault/path/to/foo?bar", "path/to/foo"),
    ],
)
def test_set(write_kv, key, exp_path, data):
    """
    Test salt.sdb.vault.set_ with current and old (question mark) syntax.
    KV v1/2 distinction is unnecessary, since that is handled in the utils module.
    """
    vault.set_(key, "super awesome")
    write_kv.assert_called_once_with(
        f"sdb://myvault/{exp_path}", data, opts=ANY, context=ANY
    )


@pytest.mark.usefixtures("write_kv_err")
def test_set_err():
    """
    Test that salt.sdb.vault.set_ raises CommandExecutionError from other exceptions
    """
    with pytest.raises(salt.exceptions.CommandExecutionError, match="damn") as exc:
        vault.set_("sdb://myvault/path/to/foo/bar", "foo")


@pytest.mark.parametrize(
    "key,exp_path",
    [
        ("sdb://myvault/path/to/foo/bar", "path/to/foo"),
        ("sdb://myvault/path/to/foo?bar", "path/to/foo"),
    ],
)
def test_get(read_kv, key, exp_path):
    """
    Test salt.sdb.vault.get_ with current and old (question mark) syntax.
    KV v1/2 distinction is unnecessary, since that is handled in the utils module.
    """
    res = vault.get(key)
    assert res == "super awesome"
    read_kv.assert_called_once_with(f"sdb://myvault/{exp_path}", opts=ANY, context=ANY)


@pytest.mark.usefixtures("read_kv")
def test_get_missing_key():
    """
    Test that salt.sdb.vault.get returns None if vault does not have the key
    but does have the entry.
    """
    res = vault.get("sdb://myvault/path/to/foo/foo")
    assert res is None


@pytest.mark.usefixtures("read_kv_not_found")
def test_get_missing():
    """
    Test that salt.sdb.vault.get returns None if vault does have the entry.
    """
    res = vault.get("sdb://myvault/path/to/foo/foo")
    assert res is None


def test_get_whole_dataset(read_kv_not_found_once, data):
    """
    Test that salt.sdb.vault.get retries the whole path without key if the
    first request reported the dataset was not found.
    """
    res = vault.get("sdb://myvault/path/to/foo")
    assert res == data
    read_kv_not_found_once.assert_called_with(
        "sdb://myvault/path/to/foo", opts=ANY, context=ANY
    )
    assert read_kv_not_found_once.call_count == 2


@pytest.mark.usefixtures("read_kv_err")
def test_get_err():
    """
    Test that salt.sdb.vault.get raises CommandExecutionError from other exceptions
    """
    with pytest.raises(salt.exceptions.CommandExecutionError, match="damn") as exc:
        vault.get("sdb://myvault/path/to/foo/bar")



"""
Dynamic DNS updates
===================

Ensure a DNS record is present or absent utilizing RFC 2136
type dynamic updates.

:depends: - `dnspython <http://www.dnspython.org/>`_

.. note::
    The ``dnspython`` module is required when managing DDNS using a TSIG key.
    If you are not using a TSIG key, DDNS is allowed by ACLs based on IP
    address and the ``dnspython`` module is not required.

Example:

.. code-block:: yaml

    webserver:
      ddns.present:
        - zone: example.com
        - ttl: 60
        - data: 111.222.333.444
        - nameserver: 123.234.345.456
        - keyfile: /srv/salt/dnspy_tsig_key.txt
"""


def __virtual__():
    if "ddns.update" in __salt__:
        return "ddns"
    return (False, "ddns module could not be loaded")


def present(name, zone, ttl, data, rdtype="A", **kwargs):
    """
    Ensures that the named DNS record is present with the given ttl.

    name
        The host portion of the DNS record, e.g., 'webserver'. Name and zone
        are concatenated when the entry is created unless name includes a
        trailing dot, so make sure that information is not duplicated in these
        two arguments.

    zone
        The zone to check/update

    ttl
        TTL for the record

    data
        Data for the DNS record. E.g., the IP address for an A record.

    rdtype
        DNS resource type. Default 'A'.

    ``**kwargs``
        Additional arguments the ddns.update function may need (e.g.
        nameserver, keyfile, keyname).  Note that the nsupdate key file can’t
        be reused by this function, the keyfile and other arguments must
        follow the `dnspython <http://www.dnspython.org/>`_ spec.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f'{rdtype} record "{name}" will be updated'
        return ret

    status = __salt__["ddns.update"](zone, name, ttl, rdtype, data, **kwargs)

    if status is None:
        ret["result"] = True
        ret["comment"] = '{} record "{}" already present with ttl of {}'.format(
            rdtype, name, ttl
        )
    elif status:
        ret["result"] = True
        ret["comment"] = f'Updated {rdtype} record for "{name}"'
        ret["changes"] = {
            "name": name,
            "zone": zone,
            "ttl": ttl,
            "rdtype": rdtype,
            "data": data,
        }
    else:
        ret["result"] = False
        ret["comment"] = 'Failed to create or update {} record for "{}"'.format(
            rdtype, name
        )
    return ret


def absent(name, zone, data=None, rdtype=None, **kwargs):
    """
    Ensures that the named DNS record is absent.

    name
        The host portion of the DNS record, e.g., 'webserver'. Name and zone
        are concatenated when the entry is created unless name includes a
        trailing dot, so make sure that information is not duplicated in these
        two arguments.

    zone
        The zone to check

    data
        Data for the DNS record. E.g., the IP address for an A record. If omitted,
        all records matching name (and rdtype, if provided) will be purged.

    rdtype
        DNS resource type. If omitted, all types will be purged.

    ``**kwargs``
        Additional arguments the ddns.update function may need (e.g.
        nameserver, keyfile, keyname).  Note that the nsupdate key file can’t
        be reused by this function, the keyfile and other arguments must
        follow the `dnspython <http://www.dnspython.org/>`_ spec.
    """
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = f'{rdtype} record "{name}" will be deleted'
        return ret

    status = __salt__["ddns.delete"](zone, name, rdtype, data, **kwargs)

    if status is None:
        ret["result"] = True
        ret["comment"] = "No matching DNS record(s) present"
    elif status:
        ret["result"] = True
        ret["comment"] = "Deleted DNS record(s)"
        ret["changes"] = {"Deleted": {"name": name, "zone": zone}}
    else:
        ret["result"] = False
        ret["comment"] = "Failed to delete DNS record(s)"
    return ret



"""
    :codeauthor: Alexander Schwartz <alexander.schwartz@gmx.net>
"""

import os

import pytest

import salt.states.archive as archive
import salt.utils.platform
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        archive: {
            "__grains__": {"os": "FooOS!"},
            "__opts__": {"cachedir": "/tmp", "test": False, "hash_type": "sha256"},
            "__env__": "test",
        }
    }


def _isfile_side_effect(path):
    """
    MagicMock side_effect for os.path.isfile(). We don't want to use dict.get
    here because we want the test to fail if there's a path we haven't
    accounted for, so that we can add it.

    NOTE: This may fall over on some platforms if /usr/bin/tar does not exist.
    If so, just add an entry in the dictionary for the path being used for tar.
    """
    if salt.utils.platform.is_windows():
        path = path.lower()
    d = {
        "/tmp/foo.tar.gz": True,
        "c:\\tmp\\foo.tar.gz": True,
        "/private/tmp/foo.tar.gz": True,
        "/tmp/out": False,
        "\\tmp\\out": False,
        "/usr/bin/tar": True,
        "/bin/tar": True,
        "/tmp/test_extracted_tar": False,
        "c:\\tmp\\test_extracted_tar": False,
        "/private/tmp/test_extracted_tar": False,
    }
    return d[path]


def test_extracted_tar():
    """
    archive.extracted tar options
    """

    if salt.utils.platform.is_windows():
        source = "C:\\tmp\\foo.tar.gz"
        tmp_dir = "C:\\tmp\\test_extracted_tar"
    elif salt.utils.platform.is_darwin():
        source = "/private/tmp/foo.tar.gz"
        tmp_dir = "/private/tmp/test_extracted_tar"
    else:
        source = "/tmp/foo.tar.gz"
        tmp_dir = "/tmp/test_extracted_tar"
    test_tar_opts = [
        "--no-anchored foo",
        "v -p --opt",
        "-v -p",
        "--long-opt -z",
        "z -v -weird-long-opt arg",
    ]
    ret_tar_opts = [
        ["tar", "xv", "--no-anchored", "foo", "-f"],
        ["tar", "xv", "-p", "--opt", "-f"],
        ["tar", "xv", "-p", "-f"],
        ["tar", "xv", "--long-opt", "-z", "-f"],
        ["tar", "xvz", "-weird-long-opt", "arg", "-f"],
    ]

    mock_true = MagicMock(return_value=True)
    mock_false = MagicMock(return_value=False)
    ret = {
        "stdout": ["cheese", "ham", "saltines"],
        "stderr": "biscuits",
        "retcode": "31337",
        "pid": "1337",
    }
    mock_run = MagicMock(return_value=ret)
    mock_source_list = MagicMock(return_value=(source, None))
    state_single_mock = MagicMock(return_value={"local": {"result": True}})
    list_mock = MagicMock(
        return_value={
            "dirs": [],
            "files": ["cheese", "saltines"],
            "links": ["ham"],
            "top_level_dirs": [],
            "top_level_files": ["cheese", "saltines"],
            "top_level_links": ["ham"],
        }
    )
    isfile_mock = MagicMock(side_effect=_isfile_side_effect)

    with patch.dict(
        archive.__opts__,
        {"test": False, "cachedir": tmp_dir, "hash_type": "sha256"},
    ), patch.dict(
        archive.__salt__,
        {
            "file.directory_exists": mock_false,
            "file.file_exists": mock_false,
            "state.single": state_single_mock,
            "file.makedirs": mock_true,
            "cmd.run_all": mock_run,
            "archive.list": list_mock,
            "file.source_list": mock_source_list,
        },
    ), patch.dict(
        archive.__states__, {"file.directory": mock_true}
    ), patch.object(
        os.path, "isfile", isfile_mock
    ), patch(
        "salt.utils.path.which", MagicMock(return_value=True)
    ):

        for test_opts, ret_opts in zip(test_tar_opts, ret_tar_opts):
            archive.extracted(
                tmp_dir, source, options=test_opts, enforce_toplevel=False
            )
            ret_opts.append(source)
            mock_run.assert_called_with(
                ret_opts, cwd=tmp_dir + os.sep, python_shell=False
            )


def test_tar_gnutar():
    """
    Tests the call of extraction with gnutar
    """
    gnutar = MagicMock(return_value="tar (GNU tar)")
    source = "/tmp/foo.tar.gz"
    mock_false = MagicMock(return_value=False)
    mock_true = MagicMock(return_value=True)
    state_single_mock = MagicMock(return_value={"local": {"result": True}})
    run_all = MagicMock(
        return_value={"retcode": 0, "stdout": "stdout", "stderr": "stderr"}
    )
    mock_source_list = MagicMock(return_value=(source, None))
    list_mock = MagicMock(
        return_value={
            "dirs": [],
            "files": ["stdout"],
            "links": [],
            "top_level_dirs": [],
            "top_level_files": ["stdout"],
            "top_level_links": [],
        }
    )
    isfile_mock = MagicMock(side_effect=_isfile_side_effect)

    with patch.dict(
        archive.__salt__,
        {
            "cmd.run": gnutar,
            "file.directory_exists": mock_false,
            "file.file_exists": mock_false,
            "state.single": state_single_mock,
            "file.makedirs": mock_true,
            "cmd.run_all": run_all,
            "archive.list": list_mock,
            "file.source_list": mock_source_list,
        },
    ), patch.dict(archive.__states__, {"file.directory": mock_true}), patch.object(
        os.path, "isfile", isfile_mock
    ), patch(
        "salt.utils.path.which", MagicMock(return_value=True)
    ):
        ret = archive.extracted(
            os.path.join(os.sep + "tmp", "out"),
            source,
            options="xvzf",
            enforce_toplevel=False,
            keep=True,
        )
        assert ret["changes"]["extracted_files"] == ["stdout"]


def test_tar_bsdtar():
    """
    Tests the call of extraction with bsdtar
    """
    bsdtar = MagicMock(return_value="tar (bsdtar)")
    source = "/tmp/foo.tar.gz"
    mock_false = MagicMock(return_value=False)
    mock_true = MagicMock(return_value=True)
    state_single_mock = MagicMock(return_value={"local": {"result": True}})
    run_all = MagicMock(
        return_value={"retcode": 0, "stdout": "stdout", "stderr": "stderr"}
    )
    mock_source_list = MagicMock(return_value=(source, None))
    list_mock = MagicMock(
        return_value={
            "dirs": [],
            "files": ["stderr"],
            "links": [],
            "top_level_dirs": [],
            "top_level_files": ["stderr"],
            "top_level_links": [],
        }
    )
    isfile_mock = MagicMock(side_effect=_isfile_side_effect)

    with patch.dict(
        archive.__salt__,
        {
            "cmd.run": bsdtar,
            "file.directory_exists": mock_false,
            "file.file_exists": mock_false,
            "state.single": state_single_mock,
            "file.makedirs": mock_true,
            "cmd.run_all": run_all,
            "archive.list": list_mock,
            "file.source_list": mock_source_list,
        },
    ), patch.dict(archive.__states__, {"file.directory": mock_true}), patch.object(
        os.path, "isfile", isfile_mock
    ), patch(
        "salt.utils.path.which", MagicMock(return_value=True)
    ):
        ret = archive.extracted(
            os.path.join(os.sep + "tmp", "out"),
            source,
            options="xvzf",
            enforce_toplevel=False,
            keep=True,
        )
        assert ret["changes"]["extracted_files"] == ["stderr"]


def test_tar_bsdtar_with_trim_output():
    """
    Tests the call of extraction with bsdtar with trim_output
    """
    bsdtar = MagicMock(return_value="tar (bsdtar)")
    source = "/tmp/foo.tar.gz"
    mock_false = MagicMock(return_value=False)
    mock_true = MagicMock(return_value=True)
    state_single_mock = MagicMock(return_value={"local": {"result": True}})
    run_all = MagicMock(
        return_value={"retcode": 0, "stdout": "stdout", "stderr": "stderr"}
    )
    mock_source_list = MagicMock(return_value=(source, None))
    list_mock = MagicMock(
        return_value={
            "dirs": [],
            "files": ["stderr"],
            "links": [],
            "top_level_dirs": [],
            "top_level_files": ["stderr"],
            "top_level_links": [],
        }
    )
    isfile_mock = MagicMock(side_effect=_isfile_side_effect)

    with patch.dict(
        archive.__salt__,
        {
            "cmd.run": bsdtar,
            "file.directory_exists": mock_false,
            "file.file_exists": mock_false,
            "state.single": state_single_mock,
            "file.makedirs": mock_true,
            "cmd.run_all": run_all,
            "archive.list": list_mock,
            "file.source_list": mock_source_list,
        },
    ), patch.dict(archive.__states__, {"file.directory": mock_true}), patch.object(
        os.path, "isfile", isfile_mock
    ), patch(
        "salt.utils.path.which", MagicMock(return_value=True)
    ):
        ret = archive.extracted(
            os.path.join(os.sep + "tmp", "out"),
            source,
            options="xvzf",
            enforce_toplevel=False,
            keep_source=True,
            trim_output=1,
        )
        assert ret["changes"]["extracted_files"] == ["stderr"]
        assert ret["comment"].endswith("Output was trimmed to 1 number of lines")


def test_extracted_when_if_missing_path_exists():
    """
    When if_missing exists, we should exit without making any changes.

    NOTE: We're not mocking the __salt__ dunder because if we actually run
    any functions from that dunder, we're doing something wrong. So, in
    those cases we'll just let it raise a KeyError and cause the test to
    fail.
    """
    name = if_missing = "/tmp/foo"
    source = "salt://foo.bar.tar"
    with patch.object(os.path, "exists", MagicMock(return_value=True)):
        ret = archive.extracted(name, source=source, if_missing=if_missing)
        assert ret["result"], ret
        assert ret["comment"] == f"Path {if_missing} exists"


def test_clean_parent_conflict():
    """
    Tests the call of extraction with gnutar with both clean_parent plus clean set to True
    """
    gnutar = MagicMock(return_value="tar (GNU tar)")
    source = "/tmp/foo.tar.gz"
    ret_comment = "Only one of 'clean' and 'clean_parent' can be set to True"
    mock_false = MagicMock(return_value=False)
    mock_true = MagicMock(return_value=True)
    state_single_mock = MagicMock(return_value={"local": {"result": True}})
    run_all = MagicMock(
        return_value={"retcode": 0, "stdout": "stdout", "stderr": "stderr"}
    )
    mock_source_list = MagicMock(return_value=(source, None))
    list_mock = MagicMock(
        return_value={
            "dirs": [],
            "files": ["stdout"],
            "links": [],
            "top_level_dirs": [],
            "top_level_files": ["stdout"],
            "top_level_links": [],
        }
    )
    isfile_mock = MagicMock(side_effect=_isfile_side_effect)

    with patch.dict(
        archive.__salt__,
        {
            "cmd.run": gnutar,
            "file.directory_exists": mock_false,
            "file.file_exists": mock_false,
            "state.single": state_single_mock,
            "file.makedirs": mock_true,
            "cmd.run_all": run_all,
            "archive.list": list_mock,
            "file.source_list": mock_source_list,
        },
    ), patch.dict(archive.__states__, {"file.directory": mock_true}), patch.object(
        os.path, "isfile", isfile_mock
    ), patch(
        "salt.utils.path.which", MagicMock(return_value=True)
    ):
        ret = archive.extracted(
            os.path.join(os.sep + "tmp", "out"),
            source,
            options="xvzf",
            enforce_toplevel=False,
            clean=True,
            clean_parent=True,
            keep=True,
        )
        assert ret["result"] is False
        assert ret["changes"] == {}
        assert ret["comment"] == ret_comment


def test_skip_files_list_verify_conflict():
    """
    Tests the call of extraction with both skip_files_list_verify and skip_verify set to True
    """
    gnutar = MagicMock(return_value="tar (GNU tar)")
    source = "/tmp/foo.tar.gz"
    ret_comment = (
        'Only one of "skip_files_list_verify" and "skip_verify" can be set to True'
    )
    mock_false = MagicMock(return_value=False)
    mock_true = MagicMock(return_value=True)
    state_single_mock = MagicMock(return_value={"local": {"result": True}})
    run_all = MagicMock(
        return_value={"retcode": 0, "stdout": "stdout", "stderr": "stderr"}
    )
    mock_source_list = MagicMock(return_value=(source, None))
    list_mock = MagicMock(
        return_value={
            "dirs": [],
            "files": ["stdout"],
            "links": [],
            "top_level_dirs": [],
            "top_level_files": ["stdout"],
            "top_level_links": [],
        }
    )
    isfile_mock = MagicMock(side_effect=_isfile_side_effect)

    with patch.dict(
        archive.__salt__,
        {
            "cmd.run": gnutar,
            "file.directory_exists": mock_false,
            "file.file_exists": mock_false,
            "state.single": state_single_mock,
            "file.makedirs": mock_true,
            "cmd.run_all": run_all,
            "archive.list": list_mock,
            "file.source_list": mock_source_list,
        },
    ), patch.dict(archive.__states__, {"file.directory": mock_true}), patch.object(
        os.path, "isfile", isfile_mock
    ), patch(
        "salt.utils.path.which", MagicMock(return_value=True)
    ):
        ret = archive.extracted(
            os.path.join(os.sep + "tmp", "out"),
            source,
            options="xvzf",
            enforce_toplevel=False,
            clean=True,
            skip_files_list_verify=True,
            skip_verify=True,
            keep=True,
        )
        assert ret["result"] is False
        assert ret["changes"] == {}
        assert ret["comment"] == ret_comment


def test_skip_files_list_verify_success():
    """
    Test that if the local and expected source hash are the same we won't do anything.
    """

    if salt.utils.platform.is_windows():
        source = "C:\\tmp\\foo.tar.gz"
        tmp_dir = "C:\\tmp\\test_extracted_tar"
    elif salt.utils.platform.is_darwin():
        source = "/private/tmp/foo.tar.gz"
        tmp_dir = "/private/tmp/test_extracted_tar"
    else:
        source = "/tmp/foo.tar.gz"
        tmp_dir = "/tmp/test_extracted_tar"

    expected_comment = (
        "Archive {} existing source sum is the same as the "
        "expected one and skip_files_list_verify argument "
        "was set to True. Extraction is not needed".format(source)
    )
    expected_ret = {
        "name": tmp_dir,
        "result": True,
        "changes": {},
        "comment": expected_comment,
    }
    mock_true = MagicMock(return_value=True)
    mock_false = MagicMock(return_value=False)
    mock_cached = MagicMock(return_value=f"{tmp_dir}/{source}")
    source_sum = {"hsum": "testhash", "hash_type": "sha256"}
    mock_hash = MagicMock(return_value=source_sum)
    mock_source_list = MagicMock(return_value=(source, None))
    isfile_mock = MagicMock(side_effect=_isfile_side_effect)

    with patch("salt.states.archive._read_cached_checksum", mock_hash):
        with patch.dict(
            archive.__opts__,
            {"test": False, "cachedir": tmp_dir, "hash_type": "sha256"},
        ), patch.dict(
            archive.__salt__,
            {
                "file.directory_exists": mock_false,
                "file.get_source_sum": mock_hash,
                "file.check_hash": mock_true,
                "cp.is_cached": mock_cached,
                "file.source_list": mock_source_list,
            },
        ), patch.object(
            os.path, "isfile", isfile_mock
        ):

            ret = archive.extracted(
                tmp_dir,
                source,
                source_hash="testhash",
                skip_files_list_verify=True,
                enforce_toplevel=False,
            )
            assert ret == expected_ret



import os

import pytest

import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.win_functions

pytestmark = [
    pytest.mark.windows_whitelisted,
]

IS_WINDOWS = salt.utils.platform.is_windows()


@pytest.mark.parametrize("test", [False, True])
def test_directory(file, tmp_path, test):
    """
    file.directory
    """
    name = tmp_path / "a_new_dir"
    ret = file.directory(name=str(name), test=test)
    if test is True:
        assert ret.result is None
        assert not name.is_dir()
    else:
        assert ret.result is True
        assert name.is_dir()


def test_directory_symlink_dry_run(file, tmp_path):
    """
    Ensure that symlinks are followed when file.directory is run with
    test=True
    """
    tmp_dir = tmp_path / "pgdata"
    sym_dir = tmp_path / "pg_data"

    tmp_dir.mkdir(0o0700)
    sym_dir.symlink_to(tmp_dir, target_is_directory=IS_WINDOWS)

    if IS_WINDOWS:
        extra_kwds = {
            "win_owner": salt.utils.win_functions.get_current_user(with_domain=False),
        }
    else:
        extra_kwds = {
            "mode": 700,
        }

    ret = file.directory(
        test=True, name=str(sym_dir), follow_symlinks=True, **extra_kwds
    )

    expected = True

    if IS_WINDOWS:
        # On Windows the result is None because there would have been changes
        # made to the directory (making Administrator the Owner)
        # https://docs.saltproject.io/en/latest/ref/states/writing.html#return-data
        expected = None

    assert ret.result is expected


def _kernel_check(lookfor):
    with salt.utils.files.fopen("/proc/version") as fp:
        versioninfo = fp.read().lower()
    return lookfor in versioninfo


@pytest.mark.skip_if_not_root
@pytest.mark.skip_on_windows(reason="Windows does not report any file modes. Skipping.")
def test_directory_max_depth(file, tmp_path):
    """
    file.directory
    Test the max_depth option by iteratively increasing the depth and
    checking that no changes deeper than max_depth have been attempted
    """

    def _get_oct_mode(name):
        """
        Return a string octal representation of the permissions for name
        """
        return salt.utils.files.normalize_mode(oct(name.stat().st_mode & 0o777))

    top = tmp_path / "top_dir"
    sub = top / "sub_dir"
    subsub = sub / "sub_sub_dir"
    dirs = [top, sub, subsub]

    initial_mode = "0111"
    changed_mode = "0555"

    # Check that we are not just running photon but the kernel matches. This
    # check should fail if we are in a photon container running on and os other
    # than photon.
    if salt.utils.platform.is_photonos() and _kernel_check("photon"):
        initial_modes = {
            0: {sub: "0750", subsub: "0110"},
            1: {sub: "0110", subsub: "0110"},
            2: {sub: "0110", subsub: "0110"},
        }
    else:
        initial_modes = {
            0: {sub: "0755", subsub: "0111"},
            1: {sub: "0111", subsub: "0111"},
            2: {sub: "0111", subsub: "0111"},
        }

    subsub.mkdir(mode=int(initial_mode, 8), exist_ok=True, parents=True)

    for depth in range(3):
        ret = file.directory(
            name=str(top),
            max_depth=depth,
            dir_mode=changed_mode,
            recurse=["mode"],
        )
        assert ret.result is True
        for changed_dir in dirs[0 : depth + 1]:
            assert changed_mode == _get_oct_mode(changed_dir)
        for untouched_dir in dirs[depth + 1 :]:
            _mode = initial_modes[depth][untouched_dir]
            assert _mode == _get_oct_mode(untouched_dir)


@pytest.mark.skip_on_windows
def test_directory_children_only(file, tmp_path):
    """
    file.directory with children_only=True
    """

    name = tmp_path / "directory_children_only_dir"
    name.mkdir(0o0700)

    strayfile = name / "strayfile"
    strayfile.touch()
    os.chmod(strayfile, 0o700)

    straydir = name / "straydir"
    straydir.mkdir(0o0700)

    # none of the children nor parent are currently set to the correct mode
    ret = file.directory(
        name=str(name),
        file_mode="0644",
        dir_mode="0755",
        recurse=["mode"],
        children_only=True,
    )
    assert ret.result is True

    # Assert parent directory's mode remains unchanged
    assert (
        oct(name.stat().st_mode)[-3:] == "700"
    ), f"Expected mode 700 for {name}, got {oct(name.stat().st_mode)[-3:]}"

    # Assert child file's mode is changed
    assert (
        oct(strayfile.stat().st_mode)[-3:] == "644"
    ), f"Expected mode 644 for {strayfile}, got {oct(strayfile.stat().st_mode)[-3:]}"

    # Assert child directory's mode is changed
    assert (
        oct(straydir.stat().st_mode)[-3:] == "755"
    ), f"Expected mode 755 for {straydir}, got {oct(straydir.stat().st_mode)[-3:]}"


def test_directory_clean(file, tmp_path):
    """
    file.directory with clean=True
    """
    name = tmp_path / "directory_clean_dir"
    name.mkdir()

    strayfile = name / "strayfile"
    strayfile.touch()

    straydir = name / "straydir"
    straydir.mkdir()

    straydir.joinpath("strayfile2").touch()

    ret = file.directory(name=str(name), clean=True)
    assert ret.result is True
    assert strayfile.exists() is False
    assert straydir.is_dir() is False
    assert name.is_dir()


def test_directory_is_idempotent(file, tmp_path):
    """
    Ensure the file.directory state produces no changes when rerun.
    """
    name = tmp_path / "a_dir_twice"

    extra_kwds = {}
    if IS_WINDOWS:
        extra_kwds["win_owner"] = salt.utils.win_functions.get_current_user(
            with_domain=True
        )

    ret = file.directory(name=str(name), **extra_kwds)
    assert ret.result is True
    assert ret.changes

    ret = file.directory(name=str(name), **extra_kwds)
    assert ret.result is True
    assert not ret.changes


@pytest.mark.parametrize("test", [False, True])
def test_directory_clean_exclude(file, tmp_path, test):
    """
    file.directory with clean=True and exclude_pat set

    Skipped on windows when test=True because clean and exclude_pat not supported
    by salt.sates.file._check_directory_win
    """
    if test is True and IS_WINDOWS:
        pytest.skip("Skipped on windows")

    name = tmp_path / "directory_clean_dir"
    name.mkdir()

    strayfile = name / "strayfile"
    strayfile.touch()

    straydir = name / "straydir"
    straydir.mkdir()

    strayfile2 = straydir / "strayfile2"
    strayfile2.touch()

    keepfile = straydir / "keepfile"
    keepfile.touch()

    exclude_pat = "E@^straydir(|/keepfile)$"
    if IS_WINDOWS:
        exclude_pat = r"E@^straydir(|\\keepfile)$"

    ret = file.directory(name=str(name), clean=True, exclude_pat=exclude_pat, test=test)
    if test is True:
        assert ret.result is None
        assert strayfile.exists()
        assert strayfile2.exists()
        assert keepfile.exists()
        assert str(strayfile) in ret.comment
        assert str(strayfile2) in ret.comment
        assert str(keepfile) not in ret.comment
    else:
        assert ret.result is True
        assert strayfile.exists() is False
        assert strayfile2.exists() is False
        assert keepfile.exists()


def test_directory_clean_require_in(modules, tmp_path, state_tree):
    """
    file.directory test with clean=True and require_in file
    """
    name = tmp_path / "a-directory"
    name.mkdir()
    good_file = name / "good-file"
    wrong_file = name / "wrong-file"
    wrong_file.write_text("foo")

    sls_contents = """
    some_dir:
      file.directory:
        - name: {name}
        - clean: true

    {good_file}:
      file.managed:
        - require_in:
          - file: some_dir
    """.format(
        name=name, good_file=good_file
    )

    with pytest.helpers.temp_file("clean-require-in.sls", sls_contents, state_tree):
        ret = modules.state.sls("clean-require-in")
        for state_run in ret:
            assert state_run.result is True

    assert good_file.exists()
    assert wrong_file.exists() is False


def test_directory_clean_require_in_with_id(modules, tmp_path, state_tree):
    """
    file.directory test with clean=True and require_in file with an ID
    different from the file name
    """
    name = tmp_path / "a-directory"
    name.mkdir()
    good_file = name / "good-file"
    wrong_file = name / "wrong-file"
    wrong_file.write_text("foo")

    sls_contents = f"""
    some_dir:
      file.directory:
        - name: {name}
        - clean: true

    some_file:
      file.managed:
        - name: {good_file}
        - require_in:
          - file: some_dir
    """

    with pytest.helpers.temp_file("clean-require-in.sls", sls_contents, state_tree):
        ret = modules.state.sls("clean-require-in")
        for state_run in ret:
            assert state_run.result is True

    assert good_file.exists()
    assert wrong_file.exists() is False


# @pytest.mark.skip_on_darwin(reaon="WAR ROOM TEMPORARY SKIP, Test is flaky on macosx")
def test_directory_clean_require_with_name(modules, tmp_path, state_tree):
    """
    file.directory test with clean=True and require with a file state
    relatively to the state's name, not its ID.
    """
    name = tmp_path / "a-directory"
    name.mkdir()
    good_file = name / "good-file"
    wrong_file = name / "wrong-file"
    wrong_file.write_text("foo")

    sls_contents = f"""
    some_dir:
      file.directory:
        - name: {name}
        - clean: true
        - require:
            # This requirement refers to the name of the following
            # state, not its ID.
            - file: {good_file}

    some_file:
      file.managed:
        - name: {good_file}
    """

    with pytest.helpers.temp_file("clean-require.sls", sls_contents, state_tree):
        ret = modules.state.sls("clean-require")
        for state_run in ret:
            assert state_run.result is True

    assert good_file.exists()
    assert wrong_file.exists() is False


def test_directory_broken_symlink(file, tmp_path):
    """
    Ensure that file.directory works even if a directory
    contains broken symbolic link
    """
    tmp_dir = tmp_path / "foo"
    tmp_dir.mkdir(0o700)
    null_file = tmp_dir / "null"
    broken_link = tmp_dir / "broken"
    broken_link.symlink_to(null_file)

    if IS_WINDOWS:
        extra_kwds = {
            "follow_symlinks": True,
            "win_owner": salt.utils.win_functions.get_current_user(with_domain=False),
        }
    else:
        extra_kwds = {
            "file_mode": 644,
            "dir_mode": 755,
        }
    ret = file.directory(name=str(tmp_dir), recurse=["mode"], **extra_kwds)
    assert ret.result is True


def _check_skip(grains):
    if grains["os"] == "MacOS":
        return True
    return False


@pytest.mark.skip_initial_gh_actions_failure(skip=_check_skip)
@pytest.mark.skip_if_not_root
@pytest.mark.skip_on_windows(
    reason="Windows fails to enforce group ownership, and test was previously skipped on windows"
)
@pytest.mark.parametrize("follow_symlinks", (True, False))
def test_issue_12209_follow_symlinks(
    file, modules, tmp_path, state_file_account, follow_symlinks
):
    """
    Ensure that symlinks are properly chowned when recursing (following
    symlinks)
    """
    # Make the directories for this test
    onedir = tmp_path / "one"
    onedir.mkdir()
    twodir = tmp_path / "two"
    twodir.symlink_to(onedir, target_is_directory=IS_WINDOWS)

    # Run the state
    ret = file.directory(
        name=str(tmp_path),
        follow_symlinks=follow_symlinks,
        user=state_file_account.username,
        group=state_file_account.group.name,
        recurse=["user", "group"],
    )
    assert ret.result is True

    if follow_symlinks:
        # Double-check, in case state mis-reported a True result. Since we are
        # following symlinks, we expect twodir to still be owned by root, but
        # onedir should be owned by the 'state_file_account.username' user.
        one_user_check = modules.file.get_user(str(onedir))
        assert one_user_check == state_file_account.username
        two_user_check = modules.file.get_user(str(twodir), follow_symlinks=False)
        assert two_user_check == "root"
        one_group_check = modules.file.get_group(str(onedir))
        assert one_group_check == state_file_account.group.name
        if salt.utils.path.which("id"):
            if "user.primary_group" in modules:
                # Which is not the case for our FreeBSD VM on our CI setup
                root_group = modules.user.primary_group("root")
                two_group_check = modules.file.get_group(
                    str(twodir), follow_symlinks=False
                )
                assert two_group_check == root_group
    else:
        # Double-check, in case state mis-reported a True result. Since we
        # are not following symlinks, we expect twodir to now be owned by
        # the 'state_file_account.username' user, just like onedir.
        one_user_check = modules.file.get_user(str(onedir))
        assert one_user_check == state_file_account.username
        two_user_check = modules.file.get_user(str(twodir), follow_symlinks=False)
        assert two_user_check == state_file_account.username
        one_group_check = modules.file.get_group(str(onedir))
        assert one_group_check == state_file_account.group.name
        two_group_check = modules.file.get_group(str(twodir), follow_symlinks=False)
        assert two_group_check == state_file_account.group.name


@pytest.mark.parametrize("backupname_isfile", [False, True])
def test_directory_backupname_force_test_mode_noclobber(
    file, tmp_path, backupname_isfile
):
    """
    Ensure that file.directory does not make changes when backupname is used
    alongside force=True and test=True.

    See https://github.com/saltstack/salt/issues/66049
    """
    source_dir = tmp_path / "source_directory"
    source_dir.mkdir()
    dest_dir = tmp_path / "dest_directory"
    backupname = tmp_path / "backup_dir"
    dest_dir.symlink_to(source_dir.resolve())

    if backupname_isfile:
        backupname.touch()
        assert backupname.is_file()

    ret = file.directory(
        name=str(dest_dir),
        allow_symlink=False,
        force=True,
        backupname=str(backupname),
        test=True,
    )

    # Confirm None result
    assert ret.result is None
    try:
        # Confirm dest_dir not modified
        assert salt.utils.path.readlink(str(dest_dir)) == str(source_dir)
    except OSError:
        pytest.fail(f"{dest_dir} was modified")

    # Confirm that comment and changes match what we expect
    assert (
        ret.comment
        == f"{dest_dir} would be backed up and replaced with a new directory"
    )
    assert ret.changes[str(dest_dir)] == {"directory": "new"}
    assert ret.changes["backup"] == f"{dest_dir} would be renamed to {backupname}"

    if backupname_isfile:
        assert ret.changes["forced"] == (
            f"Existing file at backup path {backupname} would be removed"
        )
    else:
        assert "forced" not in ret.changes



import os.path
import random
import string
import sys
from copy import deepcopy

import pytest

import salt.config
import salt.loader
import salt.modules.boto_secgroup as boto_secgroup
from salt.utils.odict import OrderedDict
from salt.utils.versions import Version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

# pylint: disable=import-error
try:
    import boto

    boto.ENDPOINTS_PATH = os.path.join(
        RUNTIME_VARS.TESTS_DIR, "unit/files/endpoints.json"
    )
    import boto.ec2  # pylint: enable=unused-import

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

try:
    from moto import mock_ec2_deprecated  # pylint: disable=no-name-in-module

    HAS_MOTO = True
except ImportError:
    HAS_MOTO = False

    def mock_ec2_deprecated(self):
        """
        if the mock_ec2_deprecated function is not available due to import failure
        this replaces the decorated function with stub_function.
        Allows boto_secgroup unit tests to use the @mock_ec2_deprecated decorator
        without a "NameError: name 'mock_ec2_deprecated' is not defined" error.
        """

        def stub_function(self):
            pass

        return stub_function


# pylint: enable=import-error


required_boto_version = "2.4.0"
vpc_id = "vpc-mjm05d27"
region = "us-east-1"
access_key = "GKTADJGHEIQSXMKKRBJ08H"
secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
conn_parameters = {
    "region": region,
    "key": access_key,
    "keyid": secret_key,
    "profile": {},
}
boto_conn_parameters = {
    "aws_access_key_id": access_key,
    "aws_secret_access_key": secret_key,
}


def _random_group_id():
    group_id = f"sg-{random.randrange(2**32):x}"
    return group_id


def _random_group_name():
    group_name = "boto_secgroup-{}".format(
        "".join((random.choice(string.ascii_lowercase)) for char in range(12))
    )
    return group_name


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    elif Version(boto.__version__) < Version(required_boto_version):
        return False
    else:
        return True


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}".format(
        required_boto_version
    ),
)
class BotoSecgroupTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.boto_secgroup module
    """

    def setup_loader_modules(self):
        opts = salt.config.DEFAULT_MASTER_OPTS.copy()
        utils = salt.loader.utils(
            opts, whitelist=["boto", "args", "systemd", "path", "platform"]
        )
        funcs = salt.loader.minion_mods(opts, utils=utils)
        return {
            boto_secgroup: {"__opts__": opts, "__utils__": utils, "__salt__": funcs}
        }

    def setUp(self):
        super().setUp()
        # __virtual__ must be caller in order for _get_conn to be injected
        boto_secgroup.__virtual__()

    def test__split_rules(self):
        """
        tests the splitting of a list of rules into individual rules
        """
        rules = [
            OrderedDict(
                [
                    ("ip_protocol", "tcp"),
                    ("from_port", 22),
                    ("to_port", 22),
                    ("grants", [OrderedDict([("cidr_ip", "0.0.0.0/0")])]),
                ]
            ),
            OrderedDict(
                [
                    ("ip_protocol", "tcp"),
                    ("from_port", 80),
                    ("to_port", 80),
                    ("grants", [OrderedDict([("cidr_ip", "0.0.0.0/0")])]),
                ]
            ),
        ]
        split_rules = [
            {
                "to_port": 22,
                "from_port": 22,
                "ip_protocol": "tcp",
                "cidr_ip": "0.0.0.0/0",
            },
            {
                "to_port": 80,
                "from_port": 80,
                "ip_protocol": "tcp",
                "cidr_ip": "0.0.0.0/0",
            },
        ]
        self.assertEqual(boto_secgroup._split_rules(rules), split_rules)

    @pytest.mark.skipif(sys.version_info >= (3, 10), reason="Fails with python 3.10")
    @mock_ec2_deprecated
    def test_create_ec2_classic(self):
        """
        Test of creation of an EC2-Classic security group. The test ensures
        that a group was created with the desired name and description
        """
        group_name = _random_group_name()
        group_description = "test_create_ec2_classic"
        boto_secgroup.create(group_name, group_description, **conn_parameters)
        conn = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        group_filter = {"group-name": group_name}
        secgroup_created_group = conn.get_all_security_groups(filters=group_filter)
        expected_create_result = [group_name, group_description, None]
        secgroup_create_result = [
            secgroup_created_group[0].name,
            secgroup_created_group[0].description,
            secgroup_created_group[0].vpc_id,
        ]
        self.assertEqual(expected_create_result, secgroup_create_result)

    @mock_ec2_deprecated
    def test_create_ec2_vpc(self):
        """
        test of creation of an EC2-VPC security group. The test ensures that a
        group was created in a given vpc with the desired name and description
        """
        group_name = _random_group_name()
        group_description = "test_create_ec2_vpc"
        # create a group using boto_secgroup
        boto_secgroup.create(
            group_name, group_description, vpc_id=vpc_id, **conn_parameters
        )
        # confirm that the group actually exists
        conn = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        group_filter = {"group-name": group_name, "vpc-id": vpc_id}
        secgroup_created_group = conn.get_all_security_groups(filters=group_filter)
        expected_create_result = [group_name, group_description, vpc_id]
        secgroup_create_result = [
            secgroup_created_group[0].name,
            secgroup_created_group[0].description,
            secgroup_created_group[0].vpc_id,
        ]
        self.assertEqual(expected_create_result, secgroup_create_result)

    @pytest.mark.skipif(sys.version_info >= (3, 10), reason="Fails with python 3.10")
    @mock_ec2_deprecated
    def test_get_group_id_ec2_classic(self):
        """
        tests that given a name of a group in EC2-Classic that the correct
        group id will be retrieved
        """
        group_name = _random_group_name()
        group_description = "test_get_group_id_ec2_classic"
        conn = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        group_classic = conn.create_security_group(
            name=group_name, description=group_description
        )
        # note that the vpc_id does not need to be created in order to create
        # a security group within the vpc when using moto
        group_vpc = conn.create_security_group(
            name=group_name, description=group_description, vpc_id=vpc_id
        )
        retrieved_group_id = boto_secgroup.get_group_id(group_name, **conn_parameters)
        self.assertEqual(group_classic.id, retrieved_group_id)

    @pytest.mark.skip(
        "test skipped because moto does not yet support group"
        " filters https://github.com/spulec/moto/issues/154",
    )
    @mock_ec2_deprecated
    def test_get_group_id_ec2_vpc(self):
        """
        tests that given a name of a group in EC2-VPC that the correct
        group id will be retrieved
        """
        group_name = _random_group_name()
        group_description = "test_get_group_id_ec2_vpc"
        conn = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        group_classic = conn.create_security_group(
            name=group_name, description=group_description
        )
        # note that the vpc_id does not need to be created in order to create
        # a security group within the vpc when using moto
        group_vpc = conn.create_security_group(
            name=group_name, description=group_description, vpc_id=vpc_id
        )
        retrieved_group_id = boto_secgroup.get_group_id(
            group_name, group_vpc, **conn_parameters
        )
        self.assertEqual(group_vpc.id, retrieved_group_id)

    @mock_ec2_deprecated
    def test_get_config_single_rule_group_name(self):
        """
        tests return of 'config' when given group name. get_config returns an OrderedDict.
        """
        group_name = _random_group_name()
        ip_protocol = "tcp"
        from_port = 22
        to_port = 22
        cidr_ip = "0.0.0.0/0"
        rules_egress = [
            {
                "to_port": None,
                "from_port": None,
                "ip_protocol": "-1",
                "cidr_ip": "0.0.0.0/0",
            }
        ]

        conn = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        group = conn.create_security_group(name=group_name, description=group_name)
        group.authorize(
            ip_protocol=ip_protocol,
            from_port=from_port,
            to_port=to_port,
            cidr_ip=cidr_ip,
        )
        # setup the expected get_config result
        expected_get_config_result = OrderedDict(
            [
                ("name", group.name),
                ("group_id", group.id),
                ("owner_id", "123456789012"),
                ("description", group.description),
                ("tags", {}),
                (
                    "rules",
                    [
                        {
                            "to_port": to_port,
                            "from_port": from_port,
                            "ip_protocol": ip_protocol,
                            "cidr_ip": cidr_ip,
                        }
                    ],
                ),
                ("rules_egress", rules_egress),
            ]
        )
        secgroup_get_config_result = boto_secgroup.get_config(
            group_id=group.id, **conn_parameters
        )
        self.assertEqual(expected_get_config_result, secgroup_get_config_result)

    @pytest.mark.skipif(sys.version_info >= (3, 10), reason="Fails with python 3.10")
    @mock_ec2_deprecated
    def test_exists_true_name_classic(self):
        """
        tests 'true' existence of a group in EC2-Classic when given name
        """
        group_name = _random_group_name()
        group_description = "test_exists_true_ec2_classic"
        conn = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        group_classic = conn.create_security_group(group_name, group_description)
        group_vpc = conn.create_security_group(
            group_name, group_description, vpc_id=vpc_id
        )
        salt_exists_result = boto_secgroup.exists(name=group_name, **conn_parameters)
        self.assertTrue(salt_exists_result)

    @mock_ec2_deprecated
    def test_exists_false_name_classic(self):
        pass

    @mock_ec2_deprecated
    def test_exists_true_name_vpc(self):
        """
        tests 'true' existence of a group in EC2-VPC when given name and vpc_id
        """
        group_name = _random_group_name()
        group_description = "test_exists_true_ec2_vpc"
        conn = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        conn.create_security_group(group_name, group_description, vpc_id=vpc_id)
        salt_exists_result = boto_secgroup.exists(
            name=group_name, vpc_id=vpc_id, **conn_parameters
        )
        self.assertTrue(salt_exists_result)

    @mock_ec2_deprecated
    def test_exists_false_name_vpc(self):
        """
        tests 'false' existence of a group in vpc when given name and vpc_id
        """
        group_name = _random_group_name()
        salt_exists_result = boto_secgroup.exists(
            group_name, vpc_id=vpc_id, **conn_parameters
        )
        self.assertFalse(salt_exists_result)

    @mock_ec2_deprecated
    def test_exists_true_group_id(self):
        """
        tests 'true' existence of a group when given group_id
        """
        group_name = _random_group_name()
        group_description = "test_exists_true_group_id"
        conn = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        group = conn.create_security_group(group_name, group_description)
        salt_exists_result = boto_secgroup.exists(group_id=group.id, **conn_parameters)
        self.assertTrue(salt_exists_result)

    @mock_ec2_deprecated
    def test_exists_false_group_id(self):
        """
        tests 'false' existence of a group when given group_id
        """
        group_id = _random_group_id()
        salt_exists_result = boto_secgroup.exists(group_id=group_id, **conn_parameters)
        self.assertFalse(salt_exists_result)

    @pytest.mark.skipif(sys.version_info >= (3, 10), reason="Fails with python 3.10")
    @mock_ec2_deprecated
    def test_delete_group_ec2_classic(self):
        """
        test deletion of a group in EC2-Classic. Test does the following:
        1. creates two groups, in EC2-Classic and one in EC2-VPC
        2. saves the group_ids to group_ids_pre_delete
        3. removes the group in EC2-VPC
        4. saves the group ids of groups to group_ids_post_delete
        5. compares the group_ids_pre_delete and group_ids_post_delete lists
           to ensure that the correct group was deleted
        """
        group_name = _random_group_name()
        group_description = "test_delete_group_ec2_classic"
        # create two groups using boto, one in EC2-Classic and one in EC2-VPC
        conn = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        group_classic = conn.create_security_group(
            name=group_name, description=group_description
        )
        group_vpc = conn.create_security_group(
            name=group_name, description=group_description, vpc_id=vpc_id
        )
        # creates a list of all the existing all_group_ids in an AWS account
        all_groups = [group.id for group in conn.get_all_security_groups()]
        # removes the EC2-Classic Security Group
        deleted = boto_secgroup.delete(name=group_name, **conn_parameters)
        expected_groups = deepcopy(all_groups)
        expected_groups.remove(group_classic.id)
        actual_groups = [group.id for group in conn.get_all_security_groups()]
        self.assertEqual(expected_groups, actual_groups)

    @mock_ec2_deprecated
    def test_delete_group_name_ec2_vpc(self):
        pass

    @mock_ec2_deprecated
    def test__get_conn_true(self):
        """
        tests ensures that _get_conn returns an boto.ec2.connection.EC2Connection object.
        """
        conn = boto.ec2.connect_to_region(region, **boto_conn_parameters)
        salt_conn = boto_secgroup._get_conn(**conn_parameters)
        self.assertEqual(conn.__class__, salt_conn.__class__)



"""
Helper functions for use by mac modules
.. versionadded:: 2016.3.0
"""

import logging
import os
import plistlib
import subprocess
import time
import xml.parsers.expat

import salt.grains.extra
import salt.modules.cmdmod
import salt.utils.args
import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
import salt.utils.timed_subprocess
from salt.exceptions import (
    CommandExecutionError,
    SaltInvocationError,
    TimedProcTimeoutError,
)

try:
    import pwd
except ImportError:
    # The pwd module is not available on all platforms
    pass


DEFAULT_SHELL = salt.grains.extra.shell()["shell"]

# Set up logging
log = logging.getLogger(__name__)

__virtualname__ = "mac_utils"

__salt__ = {
    "cmd.run_all": salt.modules.cmdmod._run_all_quiet,
    "cmd.run": salt.modules.cmdmod._run_quiet,
}


def __virtual__():
    """
    Load only on Mac OS
    """
    if not salt.utils.platform.is_darwin():
        return (
            False,
            "The mac_utils utility could not be loaded: "
            "utility only works on MacOS systems.",
        )

    return __virtualname__


def _run_all(cmd):
    """

    Args:
        cmd:

    Returns:

    """
    if not isinstance(cmd, list):
        cmd = salt.utils.args.shlex_split(cmd, posix=False)

    for idx, item in enumerate(cmd):
        if not isinstance(cmd[idx], str):
            cmd[idx] = str(cmd[idx])

    cmd = " ".join(cmd)

    run_env = os.environ.copy()

    kwargs = {
        "cwd": None,
        "shell": DEFAULT_SHELL,
        "env": run_env,
        "stdin": None,
        "stdout": subprocess.PIPE,
        "stderr": subprocess.PIPE,
        "with_communicate": True,
        "timeout": None,
        "bg": False,
    }

    try:
        proc = salt.utils.timed_subprocess.TimedProc(cmd, **kwargs)

    except OSError as exc:
        raise CommandExecutionError(
            "Unable to run command '{}' with the context '{}', reason: {}".format(
                cmd, kwargs, exc
            )
        )

    ret = {}

    try:
        proc.run()
    except TimedProcTimeoutError as exc:
        ret["stdout"] = str(exc)
        ret["stderr"] = ""
        ret["retcode"] = 1
        ret["pid"] = proc.process.pid
        return ret

    out, err = proc.stdout, proc.stderr

    if out is not None:
        out = salt.utils.stringutils.to_str(out).rstrip()
    if err is not None:
        err = salt.utils.stringutils.to_str(err).rstrip()

    ret["pid"] = proc.process.pid
    ret["retcode"] = proc.process.returncode
    ret["stdout"] = out
    ret["stderr"] = err

    return ret


def _check_launchctl_stderr(ret):
    """
    helper class to check the launchctl stderr.
    launchctl does not always return bad exit code
    if there is a failure
    """
    err = ret["stderr"].lower()
    if "service is disabled" in err:
        return True
    return False


def execute_return_success(cmd):
    """
    Executes the passed command. Returns True if successful

    :param str cmd: The command to run

    :return: True if successful, otherwise False
    :rtype: bool

    :raises: Error if command fails or is not supported
    """

    ret = _run_all(cmd)
    log.debug("Execute return success %s: %r", cmd, ret)

    if ret["retcode"] != 0 or "not supported" in ret["stdout"].lower():
        msg = f"Command Failed: {cmd}\n"
        msg += "Return Code: {}\n".format(ret["retcode"])
        msg += "Output: {}\n".format(ret["stdout"])
        msg += "Error: {}\n".format(ret["stderr"])
        raise CommandExecutionError(msg)

    return True


def execute_return_result(cmd):
    """
    Executes the passed command. Returns the standard out if successful

    :param str cmd: The command to run

    :return: The standard out of the command if successful, otherwise returns
    an error
    :rtype: str

    :raises: Error if command fails or is not supported
    """
    ret = _run_all(cmd)

    if ret["retcode"] != 0 or "not supported" in ret["stdout"].lower():
        msg = f"Command Failed: {cmd}\n"
        msg += "Return Code: {}\n".format(ret["retcode"])
        msg += "Output: {}\n".format(ret["stdout"])
        msg += "Error: {}\n".format(ret["stderr"])
        raise CommandExecutionError(msg)

    return ret["stdout"]


def parse_return(data):
    """
    Returns the data portion of a string that is colon separated.

    :param str data: The string that contains the data to be parsed. Usually the
    standard out from a command

    For example:
    ``Time Zone: America/Denver``
    will return:
    ``America/Denver``
    """

    if ": " in data:
        return data.split(": ")[1]
    if ":\n" in data:
        return data.split(":\n")[1]
    else:
        return data


def validate_enabled(enabled):
    """
    Helper function to validate the enabled parameter. Boolean values are
    converted to "on" and "off". String values are checked to make sure they are
    either "on" or "off"/"yes" or "no". Integer ``0`` will return "off". All
    other integers will return "on"

    :param enabled: Enabled can be boolean True or False, Integers, or string
    values "on" and "off"/"yes" and "no".
    :type: str, int, bool

    :return: "on" or "off" or errors
    :rtype: str
    """
    if isinstance(enabled, str):
        if enabled.lower() not in ["on", "off", "yes", "no"]:
            msg = (
                "\nMac Power: Invalid String Value for Enabled.\n"
                "String values must be 'on' or 'off'/'yes' or 'no'.\n"
                "Passed: {}".format(enabled)
            )
            raise SaltInvocationError(msg)

        return "on" if enabled.lower() in ["on", "yes"] else "off"

    return "on" if bool(enabled) else "off"


def confirm_updated(value, check_fun, normalize_ret=False, wait=5):
    """
    Wait up to ``wait`` seconds for a system parameter to be changed before
    deciding it hasn't changed.

    :param str value: The value indicating a successful change

    :param function check_fun: The function whose return is compared with
        ``value``

    :param bool normalize_ret: Whether to normalize the return from
        ``check_fun`` with ``validate_enabled``

    :param int wait: The maximum amount of seconds to wait for a system
        parameter to change
    """
    for i in range(wait):
        state = validate_enabled(check_fun()) if normalize_ret else check_fun()
        log.debug(
            "Confirm update try: %d func:%r state:%s value:%s",
            i,
            check_fun,
            state,
            value,
        )
        if value in state:
            return True
        time.sleep(1)
    return False


def launchctl(sub_cmd, *args, **kwargs):
    """
    Run a launchctl command and raise an error if it fails

    Args: additional args are passed to launchctl
        sub_cmd (str): Sub command supplied to launchctl

    Kwargs: passed to ``cmd.run_all``
        return_stdout (bool): A keyword argument. If true return the stdout of
            the launchctl command

    Returns:
        bool: ``True`` if successful
        str: The stdout of the launchctl command if requested

    Raises:
        CommandExecutionError: If command fails

    CLI Example:

    .. code-block:: bash

        import salt.utils.mac_service
        salt.utils.mac_service.launchctl('debug', 'org.cups.cupsd')
    """
    # Get return type
    return_stdout = kwargs.pop("return_stdout", False)

    # Construct command
    cmd = ["launchctl", sub_cmd]
    cmd.extend(args)

    # fix for https://github.com/saltstack/salt/issues/57436
    if sub_cmd == "bootout":
        kwargs["success_retcodes"] = [
            36,
        ]

    # Run command
    kwargs["python_shell"] = False
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    ret = __salt__["cmd.run_all"](cmd, **kwargs)
    error = _check_launchctl_stderr(ret)

    # Raise an error or return successful result
    if ret["retcode"] or error:
        out = f"Failed to {sub_cmd} service:\n"
        out += "stdout: {}\n".format(ret["stdout"])
        out += "stderr: {}\n".format(ret["stderr"])
        out += "retcode: {}".format(ret["retcode"])
        raise CommandExecutionError(out)
    else:
        return ret["stdout"] if return_stdout else True


def _read_plist_file(root, file_name):
    """
    :param root: The root path of the plist file
    :param file_name: The name of the plist file
    :return:  An empty dictionary if the plist file was invalid, otherwise, a dictionary with plist data
    """
    file_path = os.path.join(root, file_name)
    log.debug("read_plist: Gathering service info for %s", file_path)

    # Must be a plist file
    if not file_path.lower().endswith(".plist"):
        log.debug("read_plist: Not a plist file: %s", file_path)
        return {}

    # ignore broken symlinks
    if not os.path.exists(os.path.realpath(file_path)):
        log.warning("read_plist: Ignoring broken symlink: %s", file_path)
        return {}

    try:
        with salt.utils.files.fopen(file_path, "rb") as handle:
            plist = plistlib.load(handle)

    except plistlib.InvalidFileException:
        # Raised in python3 if the file is not XML.
        # There's nothing we can do; move on to the next one.
        log.warning(
            'read_plist: Unable to parse "%s" as it is invalid XML: InvalidFileException.',
            file_path,
        )
        return {}

    except ValueError as err:
        # fixes https://github.com/saltstack/salt/issues/58143
        # choosing not to log a Warning as this would happen on BigSur+ machines.
        log.debug(
            "Caught ValueError: '%s', while trying to parse '%s'.", err, file_path
        )
        return {}

    except xml.parsers.expat.ExpatError:
        # Raised by py3 if the file is XML, but with errors.
        log.warning(
            'read_plist: Unable to parse "%s" as it is invalid XML: xml.parsers.expat.ExpatError.',
            file_path,
        )
        return {}

    if "Label" not in plist:
        # not all launchd plists contain a Label key
        log.debug(
            "read_plist: Service does not contain a Label key. Skipping %s.", file_path
        )
        return {}

    return {
        "file_name": file_name,
        "file_path": file_path,
        "plist": plist,
    }


def _available_services(refresh=False):
    """
    This is a helper function for getting the available macOS services.

    The strategy is to look through the known system locations for
    launchd plist files, parse them, and use their information for
    populating the list of services. Services can run without a plist
    file present, but normally services which have an automated startup
    will have a plist file, so this is a minor compromise.
    """
    if "available_services" in __context__ and not refresh:
        log.debug("Found context for available services.")
        __context__["using_cached_services"] = True
        return __context__["available_services"]

    launchd_paths = {
        "/Library/LaunchAgents",
        "/Library/LaunchDaemons",
        "/System/Library/LaunchAgents",
        "/System/Library/LaunchDaemons",
    }

    agent_path = "/Users/{}/Library/LaunchAgents"
    launchd_paths.update(
        {
            agent_path.format(user)
            for user in os.listdir("/Users/")
            if os.path.isdir(agent_path.format(user))
        }
    )

    result = {}
    for launch_dir in launchd_paths:
        for root, dirs, files in salt.utils.path.os_walk(launch_dir):
            for file_name in files:
                data = _read_plist_file(root, file_name)
                if data:
                    result[data["plist"]["Label"].lower()] = data

    # put this in __context__ as this is a time consuming function.
    # a fix for this issue. https://github.com/saltstack/salt/issues/48414
    __context__["available_services"] = result
    # this is a fresh gathering of services, set cached to false
    __context__["using_cached_services"] = False

    return result


def available_services(refresh=False):
    """
    Return a dictionary of all available services on the system

    :param bool refresh: If you wish to refresh the available services
    as this data is cached on the first run.

    Returns:
        dict: All available services

    CLI Example:

    .. code-block:: bash

        import salt.utils.mac_service
        salt.utils.mac_service.available_services()
    """
    log.debug("Loading available services")
    return _available_services(refresh)


def console_user(username=False):
    """
    Gets the UID or Username of the current console user.

    :return: The uid or username of the console user.

    :param bool username: Whether to return the username of the console
    user instead of the UID. Defaults to False

    :rtype: Interger of the UID, or a string of the username.

    Raises:
        CommandExecutionError: If we fail to get the UID.

    CLI Example:

    .. code-block:: bash

        import salt.utils.mac_service
        salt.utils.mac_service.console_user()
    """
    try:
        # returns the 'st_uid' stat from the /dev/console file.
        uid = os.stat("/dev/console")[4]
    except (OSError, IndexError):
        # we should never get here but raise an error if so
        raise CommandExecutionError("Failed to get a UID for the console user.")

    if username:
        return pwd.getpwuid(uid)[0]

    return uid


def git_is_stub():
    """
    Return whether macOS git is the standard OS stub or a real binary.
    """
    # On a fresh macOS install, /usr/bin/git is a stub, which if
    # accessed, triggers a UI dialog box prompting the user to install
    # the developer command line tools. We don't want that! So instead,
    # running the below command will return a path to the installed dev
    # tools and retcode 0, or print a bunch of info to stderr and
    # retcode 2.
    try:
        cmd = ["/usr/bin/xcode-select", "-p"]
        _ = subprocess.check_call(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, timeout=1
        )
        log.debug("Xcode command line tools present")
        return False
    except subprocess.CalledProcessError:
        log.debug("Xcode command line tools not present")
        return True



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    Test cases for salt.modules.key
"""

import os.path

import pytest

import salt.modules.key as key
import salt.utils.crypt
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {key: {}}


def test_finger():
    """
    Test for finger
    """
    with patch.object(os.path, "join", return_value="A"):
        with patch.object(salt.utils.crypt, "pem_finger", return_value="A"):
            with patch.dict(
                key.__opts__,
                {"pki_dir": MagicMock(return_value="A"), "hash_type": "sha256"},
            ):
                assert key.finger() == "A"


def test_finger_master():
    """
    Test for finger
    """
    with patch.object(os.path, "join", return_value="A"):
        with patch.object(salt.utils.crypt, "pem_finger", return_value="A"):
            with patch.dict(key.__opts__, {"pki_dir": "A", "hash_type": "sha256"}):
                assert key.finger_master() == "A"



"""
Manage GPG keychains, add keys, create keys, retrieve keys from keyservers.
Sign, encrypt, sign plus encrypt and verify text and files.

.. versionadded:: 2015.5.0

.. note::

    The ``python-gnupg`` library and ``gpg`` binary are required to be
    installed.
    Be aware that the alternate ``gnupg`` and ``pretty-bad-protocol``
    libraries are not supported.

"""

import functools
import logging
import os
import re
import time

import salt.utils.data
import salt.utils.files
import salt.utils.immutabletypes as immutabletypes
import salt.utils.path
import salt.utils.stringutils
import salt.utils.versions
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)

try:
    import gnupg

    HAS_GPG_BINDINGS = True
except ImportError:
    HAS_GPG_BINDINGS = False


__virtualname__ = "gpg"

# Map of letters indicating key validity to pretty string (for display)
LETTER_TRUST_DICT = immutabletypes.freeze(
    {
        "e": "Expired",
        "q": "Unknown",
        "n": "Not Trusted",
        "f": "Fully Trusted",
        "m": "Marginally Trusted",
        "u": "Ultimately Trusted",
        "r": "Revoked",
        "-": "Unknown",
    }
)


# Map of allowed `trust_level` param values in `trust_key`
# to trust parameter for python-gnupg trust_keys (to manage owner trust)
TRUST_KEYS_TRUST_LEVELS = immutabletypes.freeze(
    {
        "expired": "TRUST_EXPIRED",
        "unknown": "TRUST_UNDEFINED",
        "not_trusted": "TRUST_NEVER",
        "marginally": "TRUST_MARGINAL",
        "fully": "TRUST_FULLY",
        "ultimately": "TRUST_ULTIMATE",
    }
)

# Map of allowed `trust_level` param values in `trust_key`
# to owner trust numeric values
NUM_TRUST_DICT = immutabletypes.freeze(
    {
        "expired": "1",
        "unknown": "2",
        "not_trusted": "3",
        "marginally": "4",
        "fully": "5",
        "ultimately": "6",
    }
)

# Map of owner trust numeric values to pretty string (for display)
INV_NUM_TRUST_DICT = immutabletypes.freeze(
    {
        "1": "Expired",
        "2": "Unknown",
        "3": "Not Trusted",
        "4": "Marginally",
        "5": "Fully Trusted",
        "6": "Ultimately Trusted",
    }
)

# Map of signature validity numeric values to pretty string (for display)
if not HAS_GPG_BINDINGS:
    VERIFY_TRUST_LEVELS = {}
elif salt.utils.versions.version_cmp(gnupg.__version__, "0.5.1") >= 0:
    VERIFY_TRUST_LEVELS = immutabletypes.freeze(
        {
            "0": "Expired",
            "1": "Undefined",
            "2": "Never",
            "3": "Marginal",
            "4": "Fully",
            "5": "Ultimate",
        }
    )
else:
    VERIFY_TRUST_LEVELS = immutabletypes.freeze(
        {
            "0": "Undefined",
            "1": "Never",
            "2": "Marginal",
            "3": "Fully",
            "4": "Ultimate",
        }
    )

_DEFAULT_KEY_SERVER = "keys.openpgp.org"


def _gpg():
    """
    Returns the path to the gpg binary
    """
    # Get the path to the gpg binary.
    return salt.utils.path.which("gpg")


def __virtual__():
    """
    Makes sure that python-gnupg and gpg are available.
    """
    if not _gpg():
        return (
            False,
            "The gpg execution module cannot be loaded: gpg binary is not in the path.",
        )

    return (
        __virtualname__
        if HAS_GPG_BINDINGS
        else (
            False,
            "The gpg execution module cannot be loaded; the gnupg python module is not"
            " installed.",
        )
    )


def _get_user_info(user=None):
    """
    Wrapper for user.info Salt function
    """
    if not user:
        # Get user Salt running as
        user = __salt__["config.option"]("user")

    userinfo = __salt__["user.info"](user)

    if not userinfo:
        if user == "salt":
            # Special case with `salt` user:
            # if it doesn't exist then fall back to user Salt running as
            userinfo = _get_user_info()
        else:
            raise SaltInvocationError(f"User {user} does not exist")

    return userinfo


def _get_user_gnupghome(user):
    """
    Return default GnuPG home directory path for a user
    """
    if user == "salt":
        gnupghome = os.path.join(__salt__["config.get"]("config_dir"), "gpgkeys")
    else:
        gnupghome = os.path.join(_get_user_info(user)["home"], ".gnupg")

    return gnupghome


def _restore_ownership(func):
    @functools.wraps(func)
    def func_wrapper(*args, **kwargs):
        """
        Wrap gpg function calls to fix permissions
        """
        user = kwargs.get("user")
        gnupghome = kwargs.get("gnupghome") or _get_user_gnupghome(user)
        keyring = kwargs.get("keyring")

        userinfo = _get_user_info(user)
        run_user = _get_user_info()

        if userinfo["uid"] != run_user["uid"]:
            group = None
            if os.path.exists(gnupghome):
                # Given user is different from one who runs Salt process,
                # need to fix ownership permissions for GnuPG home dir
                group = __salt__["file.gid_to_group"](run_user["gid"])
                for path in [gnupghome] + __salt__["file.find"](gnupghome):
                    __salt__["file.chown"](path, run_user["name"], group)
            if keyring and os.path.exists(keyring):
                if group is None:
                    group = __salt__["file.gid_to_group"](run_user["gid"])
                __salt__["file.chown"](keyring, run_user["name"], group)

        # Filter special kwargs
        for key in list(kwargs):
            if key.startswith("__"):
                del kwargs[key]

        ret = func(*args, **kwargs)

        if userinfo["uid"] != run_user["uid"]:
            group = __salt__["file.gid_to_group"](userinfo["gid"])
            for path in [gnupghome] + __salt__["file.find"](gnupghome):
                __salt__["file.chown"](path, user, group)
            if keyring and os.path.exists(keyring):
                __salt__["file.chown"](keyring, user, group)
        return ret

    return func_wrapper


def _create_gpg(user=None, gnupghome=None, keyring=None):
    """
    Create the GPG object
    """
    if not gnupghome:
        gnupghome = _get_user_gnupghome(user)

    if keyring and not isinstance(keyring, str):
        raise SaltInvocationError(
            "Please pass keyring as a string. Multiple keyrings are not allowed"
        )

    gpg = gnupg.GPG(gnupghome=gnupghome, keyring=keyring)

    return gpg


def _list_keys(secret=False, user=None, gnupghome=None, keyring=None):
    """
    Helper function for listing keys
    """
    gpg = _create_gpg(user=user, gnupghome=gnupghome, keyring=keyring)
    _keys = gpg.list_keys(secret)
    return _keys


def _search_keys(text, keyserver, user=None, gnupghome=None):
    """
    Helper function for searching keys from keyserver
    """
    gpg = _create_gpg(user=user, gnupghome=gnupghome)
    if keyserver:
        _keys = gpg.search_keys(text, keyserver)
    else:
        _keys = gpg.search_keys(text)
    return _keys


def search_keys(text, keyserver=None, user=None, gnupghome=None):
    """
    Search for keys on a keyserver

    text
        Text to search the keyserver for, e.g. email address, keyID or fingerprint.

    keyserver
        Keyserver to use for searching for GPG keys, defaults to keys.openpgp.org.

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.search_keys user@example.com

        salt '*' gpg.search_keys user@example.com keyserver=keyserver.ubuntu.com

        salt '*' gpg.search_keys user@example.com keyserver=keyserver.ubuntu.com user=username

    """
    if not keyserver:
        keyserver = _DEFAULT_KEY_SERVER

    _keys = []
    for _key in _search_keys(text, keyserver, user=user, gnupghome=gnupghome):
        tmp = {"keyid": _key["keyid"], "uids": _key["uids"]}

        expires = _key.get("expires", None)
        date = _key.get("date", None)
        length = _key.get("length", None)

        if expires:
            tmp["expires"] = time.strftime(
                "%Y-%m-%d", time.localtime(float(_key["expires"]))
            )
        if date:
            tmp["created"] = time.strftime(
                "%Y-%m-%d", time.localtime(float(_key["date"]))
            )
        if length:
            tmp["keyLength"] = _key["length"]
        _keys.append(tmp)
    return _keys


def list_keys(user=None, gnupghome=None, keyring=None):
    """
    List keys in GPG keychain

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.list_keys

    """
    _keys = []
    for _key in _list_keys(user=user, gnupghome=gnupghome, keyring=keyring):
        tmp = {
            "keyid": _key["keyid"],
            "fingerprint": _key["fingerprint"],
            "uids": _key["uids"],
        }

        expires = _key.get("expires", None)
        date = _key.get("date", None)
        length = _key.get("length", None)
        owner_trust = _key.get("ownertrust", None)
        trust = _key.get("trust", None)

        if expires:
            tmp["expires"] = time.strftime(
                "%Y-%m-%d", time.localtime(float(_key["expires"]))
            )
        if date:
            tmp["created"] = time.strftime(
                "%Y-%m-%d", time.localtime(float(_key["date"]))
            )
        if length:
            tmp["keyLength"] = _key["length"]
        if owner_trust:
            tmp["ownerTrust"] = LETTER_TRUST_DICT[_key["ownertrust"]]
        if trust:
            tmp["trust"] = LETTER_TRUST_DICT[_key["trust"]]
        _keys.append(tmp)
    return _keys


def list_secret_keys(user=None, gnupghome=None, keyring=None):
    """
    List secret keys in GPG keychain

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.list_secret_keys

    """
    _keys = []
    for _key in _list_keys(
        user=user, gnupghome=gnupghome, keyring=keyring, secret=True
    ):
        tmp = {
            "keyid": _key["keyid"],
            "fingerprint": _key["fingerprint"],
            "uids": _key["uids"],
        }

        expires = _key.get("expires", None)
        date = _key.get("date", None)
        length = _key.get("length", None)
        owner_trust = _key.get("ownertrust", None)
        trust = _key.get("trust", None)

        if expires:
            tmp["expires"] = time.strftime(
                "%Y-%m-%d", time.localtime(float(_key["expires"]))
            )
        if date:
            tmp["created"] = time.strftime(
                "%Y-%m-%d", time.localtime(float(_key["date"]))
            )
        if length:
            tmp["keyLength"] = _key["length"]
        if owner_trust:
            tmp["ownerTrust"] = LETTER_TRUST_DICT[_key["ownertrust"]]
        if trust:
            tmp["trust"] = LETTER_TRUST_DICT[_key["trust"]]
        _keys.append(tmp)
    return _keys


@_restore_ownership
def create_key(
    key_type="RSA",
    key_length=1024,
    name_real="Autogenerated Key",
    name_comment="Generated by SaltStack",
    name_email=None,
    subkey_type=None,
    subkey_length=None,
    expire_date=None,
    use_passphrase=False,
    user=None,
    gnupghome=None,
    keyring=None,
):
    """
    Create a key in the GPG keychain

    .. note::

        GPG key generation requires *a lot* of entropy and randomness.
        Difficult to do over a remote connection, consider having
        another process available which is generating randomness for
        the machine. Also especially difficult on virtual machines,
        consider the `rng-tools
        <http://www.gnu.org/software/hurd/user/tlecarrour/rng-tools.html>`_
        package.

        The create_key process takes awhile so increasing the timeout
        may be necessary, e.g. -t 15.

    key_type
        The type of the primary key to generate. It must be capable of signing.
        'RSA' or 'DSA'.

    key_length
        The length of the primary key in bits.

    name_real
        The real name of the user identity which is represented by the key.

    name_comment
        A comment to attach to the user id.

    name_email
        An email address for the user.

    subkey_type
        The type of the secondary key to generate.

    subkey_length
        The length of the secondary key in bits.

    expire_date
        The expiration date for the primary and any secondary key.
        You can specify an ISO date, A number of days/weeks/months/years,
        an epoch value, or 0 for a non-expiring key.

    use_passphrase
        Whether to use a passphrase with the signing key. The passphrase is
        retrieved from the Pillar key ``gpg_passphrase``.

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt -t 15 '*' gpg.create_key

    """
    ret = {"res": True, "fingerprint": "", "message": ""}

    create_params = {
        "key_type": key_type,
        "key_length": key_length,
        "name_real": name_real,
        "name_comment": name_comment,
    }

    gpg = _create_gpg(user=user, gnupghome=gnupghome, keyring=keyring)

    if name_email:
        create_params["name_email"] = name_email

    if subkey_type:
        create_params["subkey_type"] = subkey_type

    if subkey_length:
        create_params["subkey_length"] = subkey_length

    if expire_date:
        create_params["expire_date"] = expire_date

    if use_passphrase:
        gpg_passphrase = __salt__["pillar.get"]("gpg_passphrase")
        if not gpg_passphrase:
            ret["res"] = False
            ret["message"] = "gpg_passphrase not available in pillar."
            return ret
        else:
            create_params["passphrase"] = gpg_passphrase
    else:
        create_params["no_protection"] = True

    input_data = gpg.gen_key_input(**create_params)

    # This includes "%no-protection" in the input file for
    # passphraseless key generation in GnuPG >= 2.1 when the
    # python-gnupg library doesn't do that.
    if "No-Protection: True" in input_data:
        temp_data = input_data.splitlines()
        temp_data.remove("No-Protection: True")
        temp_data.insert(temp_data.index("%commit"), "%no-protection")
        input_data = "\n".join(temp_data) + "\n"

    key = gpg.gen_key(input_data)
    if key.fingerprint:
        ret["fingerprint"] = key.fingerprint
        ret["message"] = "GPG key pair successfully generated."
    else:
        ret["res"] = False
        ret["message"] = "Unable to generate GPG key pair."
    return ret


def delete_key(
    keyid=None,
    fingerprint=None,
    delete_secret=False,
    user=None,
    gnupghome=None,
    use_passphrase=True,
    keyring=None,
):
    """
    Delete a key from the GPG keychain.

    keyid
        The keyid of the key to be deleted.

    fingerprint
        The fingerprint of the key to be deleted.

    delete_secret
        Whether to delete a corresponding secret key prior to deleting the public key.
        Secret keys must be deleted before deleting any corresponding public keys.

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    use_passphrase
        Whether to use a passphrase with the signing key. The passphrase is retrieved
        from the Pillar key ``gpg_passphrase``. Note that this defaults to True here,
        contrary to the rest of the module functions that provide this parameter.

        .. versionadded:: 3003

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.delete_key keyid=3FAD9F1E

        salt '*' gpg.delete_key fingerprint=53C96788253E58416D20BCD352952C84C3252192

        salt '*' gpg.delete_key keyid=3FAD9F1E user=username

        salt '*' gpg.delete_key keyid=3FAD9F1E user=username delete_secret=True

    """
    ret = {"res": True, "message": ""}

    if fingerprint and keyid:
        ret["res"] = False
        ret["message"] = "Only specify one argument, fingerprint or keyid"
        return ret

    if not fingerprint and not keyid:
        ret["res"] = False
        ret["message"] = "Required argument, fingerprint or keyid"
        return ret

    gpg = _create_gpg(user=user, gnupghome=gnupghome, keyring=keyring)
    key = get_key(
        keyid=keyid,
        fingerprint=fingerprint,
        user=user,
        gnupghome=gnupghome,
        keyring=keyring,
    )

    def __delete_key(fingerprint, secret, use_passphrase):
        if secret and use_passphrase:
            gpg_passphrase = __salt__["pillar.get"]("gpg_passphrase")
            if not gpg_passphrase:
                return "gpg_passphrase not available in pillar."
            else:
                out = gpg.delete_keys(fingerprint, secret, passphrase=gpg_passphrase)
        else:
            out = gpg.delete_keys(fingerprint, secret, expect_passphrase=False)
        return out

    if key:
        fingerprint = key["fingerprint"]
        skey = get_secret_key(
            keyid=keyid,
            fingerprint=fingerprint,
            user=user,
            gnupghome=gnupghome,
            keyring=keyring,
        )
        if skey:
            if not delete_secret:
                ret["res"] = False
                ret["message"] = (
                    "Secret key exists, delete first or pass delete_secret=True."
                )
                return ret
            else:
                out = __delete_key(fingerprint, True, use_passphrase)
                if str(out) == "ok":
                    # Delete the secret key
                    ret["message"] = f"Secret key for {fingerprint} deleted\n"
                else:
                    ret["res"] = False
                    ret["message"] = (
                        f"Failed to delete secret key for {fingerprint}: {out}"
                    )
                    return ret

        # Delete the public key
        out = __delete_key(fingerprint, False, use_passphrase)
        if str(out) == "ok":
            ret["res"] = True
            ret["message"] += f"Public key for {fingerprint} deleted"
        else:
            ret["res"] = False
            ret["message"] += f"Failed to delete public key for {fingerprint}: {out}"
    else:
        ret["res"] = False
        ret["message"] = "Key not available in keychain."
    return ret


def get_key(keyid=None, fingerprint=None, user=None, gnupghome=None, keyring=None):
    """
    Get a key from the GPG keychain

    keyid
        The key ID (short or long) of the key to be retrieved.

    fingerprint
        The fingerprint of the key to be retrieved.

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.get_key keyid=3FAD9F1E

        salt '*' gpg.get_key fingerprint=53C96788253E58416D20BCD352952C84C3252192

        salt '*' gpg.get_key keyid=3FAD9F1E user=username

    """
    tmp = {}
    for _key in _list_keys(user=user, gnupghome=gnupghome, keyring=keyring):
        if (
            _key["fingerprint"] == fingerprint
            or _key["keyid"] == keyid
            or _key["keyid"][8:] == keyid
        ):
            tmp["keyid"] = _key["keyid"]
            tmp["fingerprint"] = _key["fingerprint"]
            tmp["uids"] = _key["uids"]

            expires = _key.get("expires", None)
            date = _key.get("date", None)
            length = _key.get("length", None)
            owner_trust = _key.get("ownertrust", None)
            trust = _key.get("trust", None)

            if expires:
                tmp["expires"] = time.strftime(
                    "%Y-%m-%d", time.localtime(float(_key["expires"]))
                )
            if date:
                tmp["created"] = time.strftime(
                    "%Y-%m-%d", time.localtime(float(_key["date"]))
                )
            if length:
                tmp["keyLength"] = _key["length"]
            if owner_trust:
                tmp["ownerTrust"] = LETTER_TRUST_DICT[_key["ownertrust"]]
            if trust:
                tmp["trust"] = LETTER_TRUST_DICT[_key["trust"]]
    if not tmp:
        return False
    else:
        return tmp


def get_secret_key(
    keyid=None, fingerprint=None, user=None, gnupghome=None, keyring=None
):
    """
    Get a secret key from the GPG keychain

    keyid
        The key ID (short or long) of the key to be retrieved.

    fingerprint
        The fingerprint of the key to be retrieved.

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.get_secret_key keyid=3FAD9F1E

        salt '*' gpg.get_secret_key fingerprint=53C96788253E58416D20BCD352952C84C3252192

        salt '*' gpg.get_secret_key keyid=3FAD9F1E user=username

    """
    tmp = {}
    for _key in _list_keys(
        user=user, gnupghome=gnupghome, keyring=keyring, secret=True
    ):
        if (
            _key["fingerprint"] == fingerprint
            or _key["keyid"] == keyid
            or _key["keyid"][8:] == keyid
        ):
            tmp["keyid"] = _key["keyid"]
            tmp["fingerprint"] = _key["fingerprint"]
            tmp["uids"] = _key["uids"]

            expires = _key.get("expires", None)
            date = _key.get("date", None)
            length = _key.get("length", None)
            owner_trust = _key.get("ownertrust", None)
            trust = _key.get("trust", None)

            if expires:
                tmp["expires"] = time.strftime(
                    "%Y-%m-%d", time.localtime(float(_key["expires"]))
                )
            if date:
                tmp["created"] = time.strftime(
                    "%Y-%m-%d", time.localtime(float(_key["date"]))
                )
            if length:
                tmp["keyLength"] = _key["length"]
            if owner_trust:
                tmp["ownerTrust"] = LETTER_TRUST_DICT[_key["ownertrust"]]
            if trust:
                tmp["trust"] = LETTER_TRUST_DICT[_key["trust"]]
    if not tmp:
        return False
    else:
        return tmp


@_restore_ownership
def import_key(text=None, filename=None, user=None, gnupghome=None, keyring=None):
    r"""
    Import a key from text or a file

    text
        The text containing the key to import.

    filename
        The path of the file containing the key to import.

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.import_key text='-----BEGIN PGP PUBLIC KEY BLOCK-----\n ... -----END PGP PUBLIC KEY BLOCK-----'
        salt '*' gpg.import_key filename='/path/to/public-key-file'

    """
    ret = {"res": True, "message": ""}

    if not text and not filename:
        raise SaltInvocationError("filename or text must be passed.")

    gpg = _create_gpg(user=user, gnupghome=gnupghome, keyring=keyring)

    if filename:
        try:
            with salt.utils.files.flopen(filename, "rb") as _fp:
                text = salt.utils.stringutils.to_unicode(_fp.read())
        except OSError:
            raise SaltInvocationError("filename does not exist.")

    imported_data = gpg.import_keys(text)

    if imported_data.imported or imported_data.imported_rsa:
        ret["message"] = "Successfully imported key(s)."
    elif imported_data.unchanged:
        ret["message"] = "Key(s) already exist in keychain."
    elif imported_data.not_imported:
        ret["res"] = False
        ret["message"] = "Unable to import key."
    elif not imported_data.count:
        ret["res"] = False
        ret["message"] = "Unable to import key."
    return ret


def export_key(
    keyids=None,
    secret=False,
    user=None,
    gnupghome=None,
    use_passphrase=False,
    output=None,
    bare=False,
    keyring=None,
):
    """
    Export a key from the GPG keychain

    keyids
        The key ID(s) of the key(s) to be exported. Can be specified as a comma
        separated string or a list. Anything which GnuPG itself accepts to identify a key
        for example, the key ID, fingerprint, user ID or email address could be used.

    secret
        Export the secret key identified by the ``keyids`` information passed.

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    use_passphrase
        Whether to use a passphrase to export the secret key.
        The passphrase is retrieved from the Pillar key ``gpg_passphrase``.

        .. versionadded:: 3003

    output
        Instead of printing to standard out, write the output to this path.

        .. versionadded:: 3006.0

    bare
        If ``True``, return the (armored) exported key block as a string without the
        standard comment/res dict.

        .. versionadded:: 3006.0

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.export_key keyids=3FAD9F1E

        salt '*' gpg.export_key keyids=3FAD9F1E secret=True

        salt '*' gpg.export_key keyids="['3FAD9F1E','3FBD8F1E']" user=username

    """
    ret = {"res": True}
    gpg = _create_gpg(user=user, gnupghome=gnupghome, keyring=keyring)

    if isinstance(keyids, str):
        keyids = keyids.split(",")

    if secret and use_passphrase:
        gpg_passphrase = __salt__["pillar.get"]("gpg_passphrase")
        if not gpg_passphrase:
            raise SaltInvocationError("gpg_passphrase not available in pillar.")
        result = gpg.export_keys(keyids, secret, passphrase=gpg_passphrase)
    else:
        result = gpg.export_keys(keyids, secret, expect_passphrase=False)

    if result and output:
        with salt.utils.files.flopen(output, "w") as fout:
            fout.write(salt.utils.stringutils.to_str(result))

    if result:
        if not bare:
            if output:
                ret["comment"] = f"Exported key data has been written to {output}"
            else:
                ret["comment"] = result
        else:
            ret = result
    else:
        if not bare:
            ret["res"] = False
        else:
            ret = False

    return ret


@_restore_ownership
def receive_keys(keyserver=None, keys=None, user=None, gnupghome=None, keyring=None):
    """
    Receive key(s) from keyserver and add them to the keychain

    keyserver
        Keyserver to use for searching for GPG keys, defaults to keys.openpgp.org

    keys
        The keyID(s) to retrieve from the keyserver. Can be specified as a comma
        separated string or a list.

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.receive_keys keys='3FAD9F1E'

        salt '*' gpg.receive_keys keys="['3FAD9F1E','3FBD9F2E']"

        salt '*' gpg.receive_keys keys=3FAD9F1E user=username

    """
    ret = {"res": True, "message": []}

    gpg = _create_gpg(user=user, gnupghome=gnupghome, keyring=keyring)

    if not keyserver:
        keyserver = _DEFAULT_KEY_SERVER

    if isinstance(keys, str):
        keys = keys.split(",")

    recv_data = gpg.recv_keys(keyserver, *keys)
    try:
        if recv_data.results:
            for result in recv_data.results:
                if "ok" in result:
                    if result["ok"] == "1":
                        ret["message"].append(
                            f"Key {result['fingerprint']} added to keychain"
                        )
                    elif result["ok"] == "0":
                        ret["message"].append(
                            f"Key {result['fingerprint']} already exists in keychain"
                        )
                elif "problem" in result:
                    ret["message"].append(
                        f"Unable to add key to keychain: {result.get('text', 'No further description')}"
                    )

        if not recv_data:
            ret["res"] = False
            ret["message"].append(f"GPG reported failure: {recv_data.stderr}")
    except AttributeError:
        ret["res"] = False
        ret["message"] = ["Invalid return from python-gpg"]

    return ret


def trust_key(
    keyid=None,
    fingerprint=None,
    trust_level=None,
    user=None,
    gnupghome=None,
    keyring=None,
):
    """
    Set the trust level for a key in the GPG keychain

    keyid
        The keyid of the key to set the trust level for.

    fingerprint
        The fingerprint of the key to set the trust level for.

    trust_level
        The trust level to set for the specified key, must be one
        of the following:
        expired, unknown, not_trusted, marginally, fully, ultimately

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

        .. versionadded:: 3007.0

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.trust_key keyid='3FAD9F1E' trust_level='marginally'
        salt '*' gpg.trust_key fingerprint='53C96788253E58416D20BCD352952C84C3252192' trust_level='not_trusted'
        salt '*' gpg.trust_key keys=3FAD9F1E trust_level='ultimately' user='username'

    """
    ret = {"res": True, "message": ""}

    if not salt.utils.data.exactly_one((keyid, fingerprint)):
        raise SaltInvocationError("Exactly one of keyid or fingerprint is required")

    if trust_level not in NUM_TRUST_DICT:
        raise SaltInvocationError(
            "ERROR: Valid trust levels - {}".format(",".join(NUM_TRUST_DICT.keys()))
        )

    key = get_key(
        keyid=keyid,
        fingerprint=fingerprint,
        user=user,
        gnupghome=gnupghome,
        keyring=keyring,
    )
    if not key:
        ret["res"] = False
        ret["message"] = f"Key {keyid or fingerprint} not in GPG keychain"
        return ret
    if not fingerprint and "fingerprint" not in key:
        ret["res"] = False
        ret["message"] = f"Fingerprint not found for keyID {keyid}"
        return ret
    fingerprint = fingerprint or key["fingerprint"]
    gpg = _create_gpg(user=user, gnupghome=gnupghome, keyring=keyring)

    try:
        res = gpg.trust_keys(fingerprint, TRUST_KEYS_TRUST_LEVELS[trust_level])
    except AttributeError:
        # python-gnupg < 0.4.2
        stdin = f"{fingerprint}:{NUM_TRUST_DICT[trust_level]}\n"
        gnupghome = gnupghome or _get_user_gnupghome(user)
        cmd = [_gpg(), "--homedir", gnupghome, "--import-ownertrust"]
        _user = user if user != "salt" else None

        if keyring:
            if not isinstance(keyring, str):
                raise SaltInvocationError(
                    "Please pass keyring as a string. Multiple keyrings are not allowed"
                )
            cmd.extend(["--no-default-keyring", "--keyring", keyring])

        res = __salt__["cmd.run_all"](cmd, stdin=stdin, runas=_user, python_shell=False)

        if not res["retcode"] == 0:
            ret["res"] = False
            ret["message"] = res["stderr"]
        else:
            if res["stderr"]:
                _match = re.findall(r"\d", res["stderr"])
                if len(_match) == 2:
                    ret["fingerprint"] = fingerprint
                    ret["message"] = "Changing ownership trust from {} to {}.".format(
                        INV_NUM_TRUST_DICT[_match[0]], INV_NUM_TRUST_DICT[_match[1]]
                    )
                else:
                    ret["fingerprint"] = fingerprint
                    ret["message"] = "Setting ownership trust to {}.".format(
                        INV_NUM_TRUST_DICT[_match[0]]
                    )
            else:
                ret["message"] = res["stderr"]
    else:
        if res.status == "ok":
            ret["res"] = True
            ret["fingerprint"] = fingerprint
            ret["message"] = "Setting ownership trust to {}.".format(
                INV_NUM_TRUST_DICT[NUM_TRUST_DICT[trust_level]]
            )
        else:
            ret["res"] = False
            ret["message"] = res.problem_reason

    return ret


def sign(
    user=None,
    keyid=None,
    text=None,
    filename=None,
    output=None,
    use_passphrase=False,
    gnupghome=None,
    keyring=None,
):
    """
    Sign a message or a file

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    keyid
        The keyid of the key to use for signing, defaults to the
        first key in the secret keyring.

    text
        The text to sign.

    filename
        The path of the file to sign.

    output
        Instead of printing to standard out, write the output to this path.

    use_passphrase
        Whether to use a passphrase with the signing key. The passphrase is
        retrieved from the Pillar key ``gpg_passphrase``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.sign text='Hello there.  How are you?'

        salt '*' gpg.sign filename='/path/to/important.file'

        salt '*' gpg.sign filename='/path/to/important.file' use_passphrase=True

    """
    if use_passphrase:
        gpg_passphrase = __salt__["pillar.get"]("gpg_passphrase")
        if not gpg_passphrase:
            raise SaltInvocationError("gpg_passphrase not available in pillar.")
    else:
        gpg_passphrase = None

    gpg = _create_gpg(user=user, gnupghome=gnupghome, keyring=keyring)

    if text:
        signed_data = gpg.sign(text, keyid=keyid, passphrase=gpg_passphrase)
    elif filename:
        with salt.utils.files.flopen(filename, "rb") as _fp:
            signed_data = gpg.sign_file(_fp, keyid=keyid, passphrase=gpg_passphrase)
        if output:
            with salt.utils.files.flopen(output, "wb") as fout:
                fout.write(salt.utils.stringutils.to_bytes(signed_data.data))
    else:
        raise SaltInvocationError("filename or text must be passed.")

    return signed_data.data


def verify(
    text=None,
    user=None,
    filename=None,
    gnupghome=None,
    signature=None,
    trustmodel=None,
    signed_by_any=None,
    signed_by_all=None,
    keyring=None,
):
    """
    Verify a message or a file

    text
        The text to verify.

    filename
        The path of the file to verify.

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    signature
        Specify the path of a detached signature.

        .. versionadded:: 2018.3.0

    trustmodel
        Explicitly define the used trust model. One of:
          - pgp
          - classic
          - tofu
          - tofu+pgp
          - direct
          - always
          - auto

        .. versionadded:: 2019.2.0

    signed_by_any
        A list of key fingerprints from which any valid signature
        will mark verification as passed. If none of the provided
        keys signed the data, verification will fail. Optional.
        Note that this does not take into account trust.

        .. versionadded:: 3007.0

    signed_by_all
        A list of key fingerprints whose signatures are required
        for verification to pass. If a single provided key did
        not sign the data, verification will fail. Optional.
        Note that this does not take into account trust.

        .. versionadded:: 3007.0

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.verify text='Hello there.  How are you?'
        salt '*' gpg.verify filename='/path/to/important.file'
        salt '*' gpg.verify filename='/path/to/important.file' trustmodel=direct

    """
    trustmodels = ("pgp", "classic", "tofu", "tofu+pgp", "direct", "always", "auto")

    if trustmodel and trustmodel not in trustmodels:
        msg = "Invalid trustmodel defined: {}. Use one of: {}".format(
            trustmodel, ", ".join(trustmodels)
        )
        log.warning(msg)
        return {"res": False, "message": msg}

    gpg = _create_gpg(user=user, gnupghome=gnupghome, keyring=keyring)
    extra_args = []

    if trustmodel:
        extra_args.extend(["--trust-model", trustmodel])

    if signed_by_any or signed_by_all:
        # batch mode stops processing on the first invalid signature.
        # This ensures all signatures are evaluated for validity.
        extra_args.append("--no-batch")
        # workaround https://github.com/vsajip/python-gnupg/issues/214
        # This issue should be fixed in versions greater than 0.5.0.
        if salt.utils.versions.version_cmp(gnupg.__version__, "0.5.0") <= 0:
            gpg.result_map["verify"] = FixedVerify

    if text:
        verified = gpg.verify(text, extra_args=extra_args)
    elif filename:
        if signature:
            # need to call with fopen instead of flopen due to:
            # https://bitbucket.org/vinay.sajip/python-gnupg/issues/76/verify_file-closes-passed-file-handle
            with salt.utils.files.fopen(signature, "rb") as _fp:
                verified = gpg.verify_file(_fp, filename, extra_args=extra_args)
        else:
            with salt.utils.files.flopen(filename, "rb") as _fp:
                verified = gpg.verify_file(_fp, extra_args=extra_args)
    else:
        raise SaltInvocationError("filename or text must be passed.")

    if not (signed_by_any or signed_by_all):
        ret = {}
        if verified.trust_level is not None:
            ret["res"] = True
            ret["username"] = verified.username
            ret["key_id"] = verified.key_id
            ret["trust_level"] = VERIFY_TRUST_LEVELS[str(verified.trust_level)]
            ret["message"] = "The signature is verified."
        else:
            ret["res"] = False
            ret["message"] = "The signature could not be verified."

        return ret

    signatures = [
        {
            "username": sig.get("username"),
            "key_id": sig["keyid"],
            "fingerprint": sig["pubkey_fingerprint"],
            "trust_level": (
                VERIFY_TRUST_LEVELS[str(sig["trust_level"])]
                if "trust_level" in sig
                else None
            ),
            "status": sig["status"],
        }
        for sig in verified.sig_info.values()
    ]
    ret = {"res": False, "message": "", "signatures": signatures}

    # be very explicit and do not default to result = True below
    any_check = all_check = False

    if signed_by_any:
        if not isinstance(signed_by_any, list):
            signed_by_any = [signed_by_any]
        any_signed = False
        for signer in signed_by_any:
            signer = str(signer)
            try:
                if any(
                    x["trust_level"] is not None and str(x["fingerprint"]) == signer
                    for x in signatures
                ):
                    any_signed = True
                    break
            except (KeyError, IndexError):
                pass

        if not any_signed:
            ret["res"] = False
            ret["message"] = (
                "None of the public keys listed in signed_by_any provided a valid signature"
            )
            return ret
        any_check = True

    if signed_by_all:
        if not isinstance(signed_by_all, list):
            signed_by_all = [signed_by_all]
        for signer in signed_by_all:
            signer = str(signer)
            try:
                if any(
                    x["trust_level"] is not None and str(x["fingerprint"]) == signer
                    for x in signatures
                ):
                    continue
            except (KeyError, IndexError):
                pass
            ret["res"] = False
            ret["message"] = (
                f"Public key {signer} has not provided a valid signature, but was listed in signed_by_all"
            )
            return ret
        all_check = True

    if bool(signed_by_any) is any_check and bool(signed_by_all) is all_check:
        ret["res"] = True
        ret["message"] = "All required keys have provided a signature"
        return ret

    ret["res"] = False
    ret["message"] = (
        "Something went wrong while checking for specific signers. This is most likely a bug"
    )
    return ret


def encrypt(
    user=None,
    recipients=None,
    text=None,
    filename=None,
    output=None,
    sign=None,
    use_passphrase=False,
    always_trust=False,
    gnupghome=None,
    bare=False,
    keyring=None,
):
    """
    Encrypt a message or a file

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    recipients
        The key ID, fingerprint, user ID or email address associated with the recipients
        key can be used.

    text
        The text to encrypt.

    filename
        The path of the file to encrypt.

    output
        Instead of printing to standard out, write the output to this path.

    sign
        Whether to sign, in addition to encrypt, the data. ``True`` to use
        default key or fingerprint to specify a different key to sign with.

    use_passphrase
        Whether to use a passphrase with the signing key.
        The passphrase is retrieved from the Pillar key ``gpg_passphrase``.

    always_trust
        Skip key validation and assume that used keys are fully trusted.

        .. versionadded:: 3006.0

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    bare
        If ``True``, return the (armored) encrypted block as a string without
        the standard comment/res dict.

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.encrypt text='Hello there.  How are you?' recipients=recipient@example.com

        salt '*' gpg.encrypt filename='/path/to/important.file' recipients=recipient@example.com

        salt '*' gpg.encrypt filename='/path/to/important.file' sign=True use_passphrase=True \\
                             recipients=recipient@example.com

    """
    ret = {"res": True, "comment": ""}
    if sign and use_passphrase:
        gpg_passphrase = __salt__["pillar.get"]("gpg_passphrase")
        if not gpg_passphrase:
            raise SaltInvocationError("gpg_passphrase not available in pillar.")
    else:
        gpg_passphrase = None

    gpg = _create_gpg(user=user, gnupghome=gnupghome, keyring=keyring)

    if text:
        result = gpg.encrypt(
            text,
            recipients,
            sign=sign,
            passphrase=gpg_passphrase,
            always_trust=always_trust,
            output=output,
        )
    elif filename:
        with salt.utils.files.flopen(filename, "rb") as _fp:
            result = gpg.encrypt_file(
                _fp,
                recipients,
                sign=sign,
                passphrase=gpg_passphrase,
                always_trust=always_trust,
                output=output,
            )
    else:
        raise SaltInvocationError("filename or text must be passed.")

    if result.ok:
        if not bare:
            if output:
                ret["comment"] = f"Encrypted data has been written to {output}"
            else:
                ret["comment"] = result.data
        else:
            ret = result.data
    else:
        if not bare:
            ret["res"] = False
            ret["comment"] = f"{result.status}.\nPlease check the salt-minion log."
        else:
            ret = False

        log.error(result.stderr)

    return ret


def decrypt(
    user=None,
    text=None,
    filename=None,
    output=None,
    use_passphrase=False,
    gnupghome=None,
    bare=False,
    keyring=None,
):
    """
    Decrypt a message or a file

    user
        Which user's keychain to access, defaults to user Salt is running as.
        Passing the user as ``salt`` will set the GnuPG home directory to
        ``/etc/salt/gpgkeys``.

    text
        The encrypted text to decrypt.

    filename
        The path of the encrypted file to decrypt.

    output
        Instead of printing to standard out, write the output to this path.

    use_passphrase
        Whether to use a passphrase with the signing key. The passphrase is retrieved
        from Pillar value ``gpg_passphrase``.

    gnupghome
        Specify the location where the GPG keyring and related files are stored.

    bare
        If ``True``, return the (armored) decrypted block as a string without the
        standard comment/res dict.

    keyring
        Limit the operation to this specific keyring, specified as
        a local filesystem path.

        .. versionadded:: 3007.0

    CLI Example:

    .. code-block:: bash

        salt '*' gpg.decrypt filename='/path/to/important.file.gpg'

        salt '*' gpg.decrypt filename='/path/to/important.file.gpg' use_passphrase=True

    """
    ret = {"res": True, "comment": ""}
    if use_passphrase:
        gpg_passphrase = __salt__["pillar.get"]("gpg_passphrase")
        if not gpg_passphrase:
            raise SaltInvocationError("gpg_passphrase not available in pillar.")
    else:
        gpg_passphrase = None

    gpg = _create_gpg(user=user, gnupghome=gnupghome, keyring=keyring)

    if text:
        result = gpg.decrypt(text, passphrase=gpg_passphrase)
    elif filename:
        with salt.utils.files.flopen(filename, "rb") as _fp:
            if output:
                result = gpg.decrypt_file(_fp, passphrase=gpg_passphrase, output=output)
            else:
                result = gpg.decrypt_file(_fp, passphrase=gpg_passphrase)
    else:
        raise SaltInvocationError("filename or text must be passed.")

    if result.ok:
        if not bare:
            if output:
                ret["comment"] = f"Decrypted data has been written to {output}"
            else:
                ret["comment"] = result.data
        else:
            ret = result.data
    else:
        if not bare:
            ret["res"] = False
            ret["comment"] = f"{result.status}.\nPlease check the salt-minion log."
        else:
            ret = False

        log.error(result.stderr)

    return ret


if HAS_GPG_BINDINGS:

    class FixedVerify(gnupg.Verify):
        """
        This is a workaround for https://github.com/vsajip/python-gnupg/issues/214.
        It ensures invalid or otherwise unverified signatures are not
        merged into sig_info in any way.

        https://github.com/vsajip/python-gnupg/commit/ee94a7ecc1a86484c9f02337e2bbdd05fd32b383
        """

        def handle_status(self, key, value):
            if "NEWSIG" == key:
                self.signature_id = None
            super().handle_status(key, value)
            if key in self.TRUST_LEVELS:
                self.signature_id = None
